# -*- coding: utf-8 -*-
"""Converter to YOLO style.

wsiprocess_to_yolo is convert helper which convert the wsiprocess output for
YOLO object detection format.
Only available for "detection" method, but works when extracting patches and
after extraction.

"""
import argparse
import random
from pathlib import Path
import json
import shutil
from collections import defaultdict

"""
'root' should be like below

./root
├── patches
│   └── class_a
│       ├── 001.png
│       ├── 002.png

│       └── 100.png
└── results.json
"""


class ToYOLOConverter:

    def __init__(self, params=False):
        if params:
            self.root = params["root"]
            self.save_to = params["save_to"]/"yolo"
            self.ratio_arg = params["ratio_arg"]
        else:
            self.getargs()
        self.parse_ratio()
        self.image_paths = defaultdict(list)
        self.read_result_file()
        self.makedirs()

    def convert(self):
        self.get_image_paths()
        self.make_images()
        self.make_labels()
        self.make_paths()

    def getargs(self):
        parser = argparse.ArgumentParser(description="wsiprocess_to_voc")
        parser.add_argument("root", type=Path)
        parser.add_argument("-s", "--save_to", default="./wp/yolo", type=Path)
        parser.add_argument("-r", "--ratio", default="8:1:1")
        args = parser.parse_args()
        self.root = args.root
        self.save_to = args.save_to
        self.ratio_arg = args.ratio

    def read_result_file(self):
        with open(self.root/"results.json", "r") as f:
            self.result_wp = json.load(f)
        self.classes = self.result_wp["classes"]
        self.filestem = Path(self.result_wp['slide']).stem
        self.image_paths = {cls: list() for cls in self.classes}

    def makedirs(self):
        if not self.save_to.exists():
            self.save_to.mkdir(parents=True)
        if not (self.save_to/"images").exists():
            (self.save_to/"images").mkdir(parents=True)
        if not (self.save_to/"labels").exists():
            (self.save_to/"labels").mkdir(parents=True)
        if not (self.save_to/"paths").exists():
            (self.save_to/"paths").mkdir(parents=True)

    def parse_ratio(self):
        self.test_is_available = len(self.ratio_arg.split(":")) == 3
        if self.test_is_available:
            train, val, test = map(int, self.ratio_arg.split(":"))
            self.ratio = {
                "train": train/(train+val+test),
                "val": val/(train+val+test),
                "test": test/(train+val+test)
            }
        else:
            train, val = map(int, self.ratio_arg.split(":"))
            self.ratio = {"train": train/(train+val), "val": val/(train+val)}

    def get_image_paths(self):
        for cls in self.classes:
            for path in (self.root/"patches"/cls).glob("*.jpg"):
                self.image_paths[cls].append(path)
            random.shuffle(self.image_paths[cls])

    def make_images(self):
        for cls in self.classes:
            for path in self.image_paths[cls]:
                filename = f"{self.filestem}_{Path(path).name}"
                shutil.copy(path, self.save_to/"images"/filename)

    def make_labels(self):
        label_dir = self.save_to/"labels"
        patch_width = self.result_wp["patch_width"]
        patch_height = self.result_wp["patch_height"]
        for patch in self.result_wp["result"]:
            bbs = []
            x = patch['x']
            y = patch['y']
            for bb in patch["bbs"]:
                label = self.classes.index(bb["class"])
                xcenter = (bb["x"] + bb["w"]/2) / patch_width
                ycenter = (bb["y"] + bb["h"]/2) / patch_height
                width = bb["w"] / patch_width
                height = bb["h"] / patch_height
                bbs.append(f"{label} {xcenter} {ycenter} {width} {height}\n")
            label_filename = f"{self.filestem}_{x:06}_{y:06}.txt"
            with open(label_dir/label_filename, "w") as f:
                f.writelines(bbs)
        with open(self.save_to/"yolo.names", "w") as f:
            f.write("\n".join(self.classes))

    def make_paths(self):
        """
        If ratio has only two parameters like "8:2", trainval.txt and test.txt
        would be generated. If ratio has 3 params like "8:1:1", trainval.txt,
        val.txt and text.txt would be generated.
        """
        train_path = f"{self.save_to}/paths/train.txt"
        test_path = f"{self.save_to}/paths/test.txt"
        if self.test_is_available:
            val_path = f"{self.save_to}/paths/val.txt"

        for cls in self.classes:
            train_count = int(len(self.image_paths[cls]) * self.ratio["train"])
            if self.test_is_available:
                val_count = int(len(self.image_paths[cls]) * self.ratio["val"])
                val_count_end = train_count+val_count

                with open(train_path, "a") as f:
                    paths = self.image_paths[cls][:train_count]
                    renamed = [self.rename_path(path) for path in paths]
                    f.writelines(renamed)
                with open(val_path, "a") as f:
                    paths = self.image_paths[cls][train_count:val_count_end]
                    renamed = [self.rename_path(path) for path in paths]
                    f.writelines(renamed)
                with open(test_path, "a") as f:
                    paths = self.image_paths[cls][val_count_end:]
                    renamed = [self.rename_path(path) for path in paths]
                    f.writelines(renamed)

            else:
                with open(train_path, "a") as f:
                    paths = self.image_paths[cls][:train_count]
                    renamed = [self.rename_path(path) for path in paths]
                    f.writelines(renamed)
                with open(test_path, "a") as f:
                    paths = self.image_paths[cls][train_count:]
                    renamed = [self.rename_path(path) for path in paths]
                    f.writelines(renamed)

    def rename_path(self, path):
        images_dir = self.save_to/"images"
        return str(images_dir/f"{self.filestem}_{Path(path).name}\n")


if __name__ == '__main__':
    converter = ToYOLOConverter()
    converter.convert()



# -*- coding: utf-8 -*-
import numpy as np
from collections import defaultdict
import sqlite3

from wsiprocess.error import AnnotationLabelError

from .parser_utils import BaseParser


class SlideRunnerAnnotation(BaseParser):
    """Annotation Parser for SlideRunner v1.31.0

    Args:
        path (str): Path to the annotation file.
        slide (wsiprocess.Slide): Slide object.

    Attributes:
        path (str): Path to the annotation file.
        classes (list): List of classes defined with ASAP.
        mask_coords (dict): Coordinates of the masks.
    """

    def __init__(self, path, slide):
        super().__init__(path)

        self.mask_coords = defaultdict(list)
        with sqlite3.connect(path) as con:
            self.cursor = con.cursor()

            self.read_classes()
            self.read_slides()
            self.read_annotations(slide.filename)
            self.read_labels()
            self.read_coordinates(slide.filename)

        self.verify_annotations()
        self.parse_mask_coords()

    def read_classes(self):
        """Read classes from Classes table.
        """
        query = "select uid, name from Classes"
        self.cursor.execute(query)
        self.classes = list()
        self.uid2cls = dict()
        for uid, name in self.cursor.fetchall():
            self.classes.append(name)
            self.uid2cls[uid] = name

    def read_slides(self):
        """Read slides from Slides table.
        A slide has unique ID and filename.
        """
        self.cursor.execute("select uid, filename from Slides")
        self.slidename2uid = {
            filename: uid
            for uid, filename in self.cursor.fetchall()
        }

    def read_annotations(self, slidename):
        """Read annotations from Annotations table.

        A annotation has unique ID, Annotation Type, and Class.

        Args:
            slidename (str): Name of the slide to filter the data.
        """
        query = "select uid, type, agreedClass from Annotations"
        query += f" where slide=='{self.slidename2uid[slidename]}'"
        self.cursor.execute(query)
        self.annotations = {
            uid: {"type": annoType, "class": cls}
            for uid, annoType, cls in self.cursor.fetchall()
        }

    def read_labels(self):
        """Read labels from Annotations_label table.

        A label has Annotation ID (sharing with Annotations), and Class.
        """
        query = "select class, annoId from Annotations_label"
        self.cursor.execute(query)
        self.labels = {
            annoId: cls
            for cls, annoId in self.cursor.fetchall()
        }

    def read_coordinates(self, slidename):
        """Read coordinates from Annotations_coordinates table.

        A coordinate has coordinate of x and y, Annotation ID and the order
        of the coordinate.

        Args:
            slidename (str): Name of the slide to filter the data.
        """
        query = "select coordinateX, coordinateY, annoId, orderIdx"
        query += " from Annotations_coordinates"
        query += f" where slide=='{self.slidename2uid[slidename]}'"
        self.cursor.execute(query)
        self.coordinates = defaultdict(dict)
        for x, y, annoId, order in self.cursor.fetchall():
            self.coordinates[annoId][order] = {"x": x, "y": y}

    def verify_annotations(self):
        """Verify the number of annotations.

        The numbers of annotations and labels and coordinates must be same.
        """
        annotation_counts = len(self.annotations)
        if not len(self.labels) == len(self.coordinates) == annotation_counts:
            raise AnnotationLabelError("Some annotations have no label.")

    def parse_mask_coords(self):
        """Parse the coordinates of the mask.

        Parse the coordinates to make a list containing lists of coordinates.
        """
        for annoId, value in self.coordinates.items():
            cls = self.uid2cls[self.labels[annoId]]

            if self.annotations[annoId]["type"] in [1, 2, 3, 4]:
                # type1 is dot annotation
                # type2 is rectangle annotation with lefttop and rightbottom
                # type3 is polygon annotation or magicwand annotation.
                # type4 is important position annotation with a dot.
                coordinate = [
                    [coord["x"], coord["y"]]
                    for order, coord in value.items()
                ]
                self.mask_coords[cls].append(coordinate)

            elif self.annotations[annoId]["type"] == 5:
                # type5 is circle anntoation.
                left = value[1]["x"]
                right = value[2]["x"]
                top = value[1]["y"]
                bottom = value[2]["y"]
                coordinate = self.bbox_to_circle(left, top, right, bottom)
                self.mask_coords[cls].append(coordinate)

            else:
                raise NotImplementedError("Unknown annotation type")

    def bbox_to_circle(self, left, top, right, bottom):
        """Convert coordinates of bounding box to circle.

        Args:
            left (int): Smaller coordinate of in the x-axis direction
            top (int): Smaller coordinate of in the y-axis direction
            right (int): Larger coordinate of in the x-axis direction
            bottom (int): Larger coordinate of in the y-axis direction

        Returns:
            coordinate (list(list)): List of list which contains one
                coordinate representing a circle in coordinates.
        """
        center_x = (left + right) / 2
        center_y = (top + bottom) / 2
        radius = np.abs((right - left) / 2)
        coordinate = [
            [
                int(center_x + np.cos(theta) * radius),
                int(center_y + np.sin(theta) * radius)
            ]
            for theta in np.linspace(0, 2*np.pi, int(radius*2))
        ]
        return coordinate



from .slide import Slide as slide
from .patcher import Patcher as patcher
from .annotation import Annotation as annotation
from .rule import Rule as rule
from .converter import Converter as converter

__version__ = "1.1.1"



# -*- coding: utf-8 -*-
"""Object to define rules for extracting patches.
Rule file should be a json file or a dict. The content of rule.json is like
below.

Example:
    Json data below defines

    - extract the patches of benign and malignant
    - benign includes stroma but excludes malignant or uncertain
    - malignant means malignant itself but excludes benign

    .. code-block:: python

        {
            "benign" : {
                "includes" : [
                    "stroma"
                ],
                "excludes : [
                    "malignant",
                    "uncertain"
                ]
            },
            "malignant": {
                "includes" : [
                ],
                "excludes" :[
                    "benign"
                ]
            }
        }
"""

from pathlib import Path
import json


class Rule:
    """Base class for rule.

    Args:
        rule(str or dict): Path to the rule.json or the rule dict.

    Attributes:
        classes (list): List of the classes. i.e. ["benign", "malignant"]
    """

    def __init__(self, rule):
        self.classes = []
        if isinstance(rule, (str, Path)):
            with open(rule, "r") as f:
                base_rule = json.load(f)

        elif isinstance(rule, dict):
            base_rule = rule

        else:
            raise NotImplementedError(
                f"rule must be str or dist, but got {type(rule)}")

        self.load_rule(base_rule)

    def load_rule(self, rule: dict):
        """Read the rule file.

        Parse the rule file and save as the classes.
        """
        for base, incl_excl in rule.items():
            self.assert_incl_excl(incl_excl)
            setattr(self, base, incl_excl)
            self.classes.append(base)

    def assert_incl_excl(self, incl_excl: dict):
        """Assert the rule has assumed keys.

        Supposed shape is like below

        {"includes": ["stroma"], "excludes": ["malignant", "uncertain"]}

        """
        assert isinstance(incl_excl, dict), "each object must be dict"
        msg = "each dict must have {}"
        assert "includes" in incl_excl.keys(), msg.format("includes")
        assert "excludes" in incl_excl.keys(), msg.format("excludes")

    def __getitem__(self, class_name):
        return getattr(self, class_name)



# -*- coding: utf-8 -*-
"""Converter to COCO style.

wsiprocess_to_coco is convert helper which convert the wsiprocess output for
COCO object detection format.
Only available for "detection" method, but works when extracting patches and
after extraction.

"""
import argparse
import shutil
import random
import json
from datetime import datetime
from pathlib import Path

import cv2


class SaveTo(type(Path())):

    def __init__(self, path):
        self.annotations = self/"annotations"
        self.train = self/"train2014"
        self.val = self/"val2014"
        self.test = self/"test2014"

        self.mkdir(exist_ok=True)
        self.annotations.mkdir(exist_ok=True)
        self.train.mkdir(exist_ok=True)
        self.val.mkdir(exist_ok=True)
        self.test.mkdir(exist_ok=True)


class ToCOCOConverter:
    """Converter class.
    """

    def __init__(self, params=False, per_wsi=False, annotation=False):
        if params:
            self.root = Path(params["root"])
            self.save_to = SaveTo(f"{params['save_to']}/coco")
            self.ratio_arg = params["ratio_arg"]
        else:
            self.getargs()
        self.per_wsi = per_wsi
        self.test_is_available = len(self.ratio_arg.split(":")) == 3
        self.phases = ["train", "val"]
        if self.test_is_available:
            self.phases.append("test")

        self.now = datetime.now().strftime('%Y/%m/%d_%H:%M:%S')
        self.paths = {phase: [] for phase in self.phases}
        self.get_base()
        self.get_ratio()
        if annotation:
            self.annotation = annotation
            self.assert_annotation()
        else:
            self.read_annotation()
        self.set_id()

    def assert_annotation(self):
        params = [
            "classes",
            "slide",
            "wsi_height",
            "wsi_width",
            "patch_height",
            "patch_width",
            "result"
        ]
        asserted = True
        for param in params:
            if not self.annotation.get(param):
                print(f"{param} not in annotation")
                asserted = False
        assert asserted, "Param not properly set"

    def set_id(self):
        last_image_id, last_annotation_id = 0, 0
        for phase in self.phases[::-1]:
            target = f"instances_{phase}2014.json"
            if (self.save_to.annotations/target).exists():
                with open(self.save_to.annotations/target, "r") as f:
                    data = json.load(f)
                    try:
                        last_image_id = data["images"][-1]["id"]
                        last_annotation_id = data["annotations"][-1]["id"]
                    except IndexError:
                        continue

        self.image_id = last_image_id + 1
        self.annotation_id = last_annotation_id + 1

    def convert(self):
        self.make_link_to_images()
        self.add_info()
        self.add_categories()
        self.add_images_and_annotations()
        self.save_data()

    def getargs(self):
        parser = argparse.ArgumentParser()
        parser.add_argument(
            "root", type=Path, help="Parent of results.json located")
        parser.add_argument("-s", "--save_to", default="./wp/coco", type=Path)
        parser.add_argument(
            "-r", "--ratio", default="8:1:1",
            help="Ratio of train and val and test ie. 8:1:1")
        args = parser.parse_args()
        self.root = args.root
        self.save_to = args.save_to
        self.ratio_arg = args.ratio

    def get_ratio(self):
        if self.per_wsi:
            self.ratio = {p: r for p, r in zip(self.phases, [0, 0, 1])}
            return
        ratios = list(map(float, self.ratio_arg.split(":")))
        self.ratio = {p: r / sum(ratios) for p, r in zip(self.phases, ratios)}

    def add_info(self):
        for phase in self.phases:
            self.coco_data[phase]["info"] = info(self.now)

    def make_link_to_images(self):
        for cls in self.annotation["classes"]:
            paths = list((self.root/"patches").glob(f"{cls}/*"))
            random.shuffle(paths)

            train_count = int(len(paths)*self.ratio["train"])
            self.paths["train"] += paths[:train_count]
            if self.test_is_available:
                val_count = int(len(paths)*self.ratio["val"])
                val_count_end = train_count + val_count
                self.paths["val"] += paths[train_count:val_count_end]
                self.paths["test"] += paths[train_count+val_count:]
            else:
                self.paths["val"] += paths[train_count:]

        for phase in self.phases:
            for image_path in self.paths[phase]:
                img_name = image_path.name
                if not (self.save_to/f"{phase}2014"/img_name).exists():
                    shutil.copy(image_path, self.save_to/f"{phase}2014")

    def get_base(self):
        self.coco_data = {}
        for phase in self.phases:
            save_to = self.save_to.annotations
            json_path = save_to/f"instances_{phase}2014.json"
            if json_path.exists():
                with open(json_path, "r") as f:
                    self.coco_data[phase] = json.load(f)
            else:
                self.coco_data[phase] = base()

    def read_annotation(self):
        with open(self.root/"results.json", "r") as f:
            self.annotation = json.load(f)

    def add_categories(self):
        for idx, cls in enumerate(self.annotation["classes"]):
            for phase in self.phases:
                self.add_category(phase, idx, cls)

    def add_images_and_annotations(self):
        if self.per_wsi:
            self.add_image("test", self.annotation["slide"])
            for path in self.paths["test"]:
                self.add_annotation("test", Path("0_0"))
            return
        for phase in self.phases:
            for path in self.paths[phase]:
                self.add_image(phase, path)
                self.add_annotation(phase, path)
                self.image_id += 1

    def add_category(self, phase, category_idx, name):
        cat = category(category_idx, name)
        if cat not in self.coco_data[phase]["categories"]:
            self.coco_data[phase]["categories"].append(cat)

    def add_image(self, phase, file_name):
        if self.per_wsi:
            width = self.annotation["wsi_width"]
            height = self.annotation["wsi_height"]
        else:
            width = self.annotation["patch_width"]
            height = self.annotation["patch_height"]
        self.coco_data[phase]["images"].append(
            image(file_name, width, height, self.image_id)
        )

    def add_annotation(self, phase, path):
        x, y = map(int, path.stem.split("_")[-2:])
        for result in self.annotation["result"]:
            if (not self.per_wsi) and (result["x"] != x or result["y"] != y):
                continue
            offsetx = x
            offsety = y
            if result.get("bbs"):
                for bb in result["bbs"]:
                    if not self.per_wsi:
                        bb["x"] %= self.annotation["patch_width"]
                        bb["y"] %= self.annotation["patch_height"]
                    self.coco_data[phase]["annotations"].append(
                        annotation(
                            bb, offsetx, offsety, self.image_id,
                            self.annotation_id, self.annotation["classes"],
                            mode="detection", per_wsi=self.per_wsi)
                    )
                    self.annotation_id += 1
            elif result.get("masks"):
                for mask in result["masks"]:
                    annotations = annotation(
                        mask, offsetx, offsety, self.image_id,
                        self.annotation_id, self.annotation["classes"],
                        mode="segmentation", per_wsi=self.per_wsi)
                    self.coco_data[phase]["annotations"].extend(annotations)
                    self.annotation_id += len(annotations)

    def save_data(self):
        for phase in self.phases:
            target = f"instances_{phase}2014.json"
            with open(self.save_to.annotations/target, "w") as f:
                json.dump(self.coco_data[phase], f, indent=4)


def base():
    return {
        "info": {},
        "licenses": [],
        "categories": [],
        "images": [],
        "annotations": []
    }


def info(now):
    return {
        "description": "wsiprocess",
        "url": "",
        "version": "1.0",
        "year": 2014,
        "contributor": "",
        "date_created": now
    }


def category(category_idx, name):
    # category_idx starts from 0 -> plus 1 to set as id
    return {
        "supercategory": "",
        "id": category_idx+1,
        "name": name}


def image(file_name, width, height, image_id):
    return {
        "license": "",
        "file_name": str(file_name),
        "coco_url": "",
        "width": width,
        "height": height,
        "date_captured": "",
        "flickr_url": "",
        "id": image_id
    }


def annotation(
        result, offsetx, offsety, image_id, annotation_id, classes, mode,
        per_wsi=False):
    if mode == "detection":
        bb = result
        return {
            "segmentation": [[
                bb["x"], bb["y"],
                bb["x"] + bb["w"], bb["y"],
                bb["x"] + bb["w"], bb["y"] + bb["h"],
                bb["x"], bb["y"] + bb["h"]
            ]],
            "area": bb["w"]*bb["h"],
            "iscrowd": 0,
            "image_id": image_id,
            "bbox": [bb["x"], bb["y"], bb["w"], bb["h"]],
            "category_id": classes.index(bb["class"]) + 1,
            "id": annotation_id
        }
    elif mode == "segmentation":
        mask = result
        mask_img = cv2.imread(mask["coords"], 0)
        contours, _ = cv2.findContours(
            mask_img, cv2.RETR_TREE, cv2.CHAIN_APPROX_SIMPLE)
        annotations = []
        for contour in contours:
            contour = contour.reshape(-1, 2)
            if per_wsi:
                contour += [offsetx, offsety]
            minx, miny = list(map(int, contour.min(axis=0)))
            maxx, maxy = list(map(int, contour.max(axis=0)))
            annotations.append({
                "segmentation": contour.ravel().tolist(),
                "area": int((mask_img == 1).sum()),
                "iscrowd": 0,
                "image_id": image_id,
                "bbox": [minx, miny, maxx - minx, maxy - miny],
                "category_id": classes.index(mask["class"]) + 1,
                "id": annotation_id
            })
        return annotations


if __name__ == '__main__':
    converter = ToCOCOConverter()
    converter.convert()



import json

from .parser_utils import BaseParser


class GeoJsonAnnotation(BaseParser):
    """Parser for GeoJson styled Annotation.

    Example:
        >>> with open("xxx.geojson", "r") as f:
        >>>     annotations = geojson.load(f)
        >>> print(annotations)
        >>> {
        >>>     "type": "FeatureCollection",
        >>>     "features": [
        >>>         {
        >>>             "type": "Feature",
        >>>             "properties": {
        >>>                 "class": 0
        >>>             },
        >>>             "geometry": {
        >>>                 "type": "Polygon",
        >>>                 "coordinates": [
        >>>                     [[0.0, 0.0], [0.0, 1.0], [1.0, 1.0], [1.0, 0.0], [0.0, 0.0]]
        >>>                 ]
        >>>             }
        >>>         }
        >>>     ]
        >>> }
    """

    def __init__(self, path):
        """
        Args:
            path (str): Path to the annotation file.
        """
        super().__init__(path)
        self.read_annotation()

    def read_annotation(self):
        """Read the annotation file."""

        with open(self.path, "r") as f:
            annotations = json.load(f)
        assert annotations["type"] == "FeatureCollection"

        classes = set()
        for annotation in annotations["features"]:
            class_ = annotation["properties"]["class"]
            classes.add(class_)
            self.mask_coords[class_].append(annotation["geometry"]["coordinates"])

        self.classes = list(classes)



# -*- coding: utf-8 -*-
"""Slide object to pass to annotation object and patcher object.
Slide is whole slide image, scanned with whole slide scanners.
Mannually you can make pyramidical tiff file, which you can handle just the
same as the scanned digital data, except for the magnification.
"""
from .error import SlideLoadError
from pathlib import Path
import numpy as np
from PIL import Image
import openslide


class Slide:
    """Slide object.

    Args:
        path (str): Path to the whole slide image file.
        backend (str): Openslide or pyivps.

    Attributes:
        path (str): Path to the whole slide image file.
        slide (pyvips.Image): pyvips Image object.
        wsi_width (int): Width of slide.
        wsi_height (int): Height of slide.
    """

    def __init__(self, path, backend="openslide"):
        self.path = path
        if not Path(path).exists():
            raise SlideLoadError("Slide File {} Not Found".format(path))
        self.filestem = Path(path).stem
        self.filename = Path(path).name
        self.backend = backend

        self.load_slide()

    def load_slide(self):
        if self.backend == "openslide":
            self.slide = openslide.OpenSlide(self.path)
            self.width, self.height = self.slide.dimensions
        elif self.backend == "pyvips":
            try:
                import pyvips
            except ImportError:
                raise ImportError("pyvips not installed")
            self.slide = pyvips.Image.new_from_file(self.path)

            self.width = self.slide.width
            self.height = self.slide.height
        else:
            raise NotImplementedError(
                "backend={} is not available".format(self.backend))
        self.set_properties()

    def __str__(self):
        return "wsiprocess.slide.Slide {} {}x{}".format(
            self.path, self.wsi_width, self.wsi_height)

    def export_thumbnail(self, save_as="./thumb.png", size=500):
        """Export thumbnail image.

        Args:
            save_as (str): Path to save as the thumbnail image.
            size (int, optional): Size of the exported thumbnail.
        """
        thumb = self.get_thumbnail(size)
        if self.backend == "openslide":
            thumb.save(save_as)
        elif self.backend == "pyvips":
            thumb.pngsave(save_as)

    def get_thumbnail(self, size=500):
        """Get thumbnail image.

        Args:
            size (int or tuple, optional): Size of the exported thumbnail.
        """
        if self.backend == "openslide":
            if isinstance(size, int):
                size = (size, size)
            return self.slide.get_thumbnail(size)
        elif self.backend == "pyvips":
            return self.slide.thumbnail_image(size)

    def set_properties(self):
        """Read the properties and set as attributes of slide obj.

        Attributes:
            magnification (int): Objective power of slide obj.
        """
        if self.backend == "openslide":
            properties = self.slide.properties
        elif self.backend == "pyvips":
            properties = {}
            for field in self.slide.get_fields():
                properties[field] = self.slide.get(field)
        for field in properties:
            setattr(self, field, properties.get(field))

        self.magnification = properties.get("openslide.objective-power")
        if self.magnification:
            self.magnification = int(self.magnification)

    def crop(self, x, y, w, h):
        if self.backend == "openslide":
            return self.slide.read_region((x, y), 0, (w, h))
        elif self.backend == "pyvips":
            patch = self.slide.crop(x, y, w, h)
            patch = np.ndarray(
                buffer=patch.write_to_memory(),
                dtype=np.uint8,
                shape=[patch.height, patch.width, patch.bands])
            return Image.fromarray(patch)



import json
from collections import defaultdict
from wsiprocess.error import AnnotationLabelError

from .parser_utils import BaseParser


class QuPathAnnotation(BaseParser):
    """Annotation Parser for QuPath.

    Args:
        path (str): Path to the annotation file.

    Attributes:
        path (str): Path to the annotation file.
        annotations (list): List of etree Elements.
        classes (list): List of classes defined with QuPath.
        mask_coords (dict): Coordinates of the masks.
    """

    def __init__(self, path):
        super().__init__(path)

        with open(path, "r") as f:
            self.annotations = json.load(f)
        self.mask_coords = defaultdict(list)

        self.read_classes()
        self.read_coordinates()

    def read_classes(self):
        classes = list()
        for annotation in self.annotations:
            if "classification" not in annotation["properties"]:
                raise AnnotationLabelError("Some annotations have no label.")
            classes.append(annotation["properties"]["classification"]["name"])
        self.classes = list(set(classes))

    def read_coordinates(self):
        """Parse the coordinates of the mask.
        """
        for annotation in self.annotations:
            cls = annotation["properties"]["classification"]["name"]
            annotation_type = annotation["geometry"]["type"]

            if annotation_type in ["Polygon"]:
                coordinates = annotation["geometry"]["coordinates"][0]

            elif annotation_type in ["LineString"]:
                coordinates = annotation["geometry"]["coordinates"]

            else:
                raise NotImplementedError(f"Unknown type {annotation_type}")

            coordinates = self.all_as_int(coordinates)
            self.mask_coords[cls].append(coordinates)

    def all_as_int(self, points):
        int_points = [[round(point[0]), round(point[1])] for point in points]
        return int_points



# -*- coding: utf-8 -*-
from typing import List
from dataclasses import dataclass
import math

from lxml import etree

from .parser_utils import BaseParser


@dataclass
class Coord:
    x: float
    y: float


@dataclass
class MPP:
    x: float
    y: float
    r: float

    def meter2px(self, meter: float, axis: str, coordformat: str) -> float:
        """Convert meter to pixel.

        Args:
            meter (float): Length in meter.
            axis (str): mpp to use along this axis.

        Returns:
            int: Length in pixel.
        """

        if coordformat == "nanometers":
            unit = 1000
        elif coordformat == "micrometers":
            unit = 1
        elif coordformat == "millimeters":
            unit = 0.001
        else:
            raise ValueError(f"Invalid coordformat: {coordformat}")
        return int(meter / unit / getattr(self, axis))


class Annotation:
    """Annotations of NDP View.

    Args:
        element (etree.Element): Element of the annotation.
        origin (Coord): Origin of the slide.
        mpp (MPP): MPP of the slide.

    Attributes:
        element (etree.Element): Element of the annotation.
        title (str): Title of the annotation. Used as the class.
        coordformat (str): Coordinate format.
        type (str): Type of the annotation. One of AnnotateCircle, AnnotateFreehandLine, AnnotateRectangle,
            AnnotatePin, AnnotateFreehand
    """

    def __init__(self, element: etree.Element, origin: Coord, mpp: MPP):
        self.element = element
        self.origin = origin
        self.mpp = mpp

        self.coordformat = element.find("coordformat").text
        self.type = element.find("annotation").attrib["displayname"]
        assert self.type in [
            "AnnotateCircle",
            "AnnotateFreehandLine",
            "AnnotateRectangle",
            "AnnotatePin",
            "AnnotateFreehand"], f"Annotation type {self.type} is not supported."

        self._coords = self.get_coords()

    @property
    def title(self):
        title = self.element.find("title").text
        return title if title else "NOTITLE"

    @property
    def coords(self):
        return self._coords

    def get_coords(self):
        if self.type == "AnnotateCircle":
            return self.get_circle_coords()
        elif self.type == "AnnotatePin":
            return self.get_pin_coords()
        else:
            return self.get_polygon_coords()

    def get_circle_coords(self) -> List[int]:
        """Get the coordinates of the circle annotation."""
        center = Coord(
            x=self.origin.x + self.mpp.meter2px(
                int(self.element.find("annotation/x").text), "x",
                self.coordformat),
            y=self.origin.y + self.mpp.meter2px(
                int(self.element.find("annotation/y").text), "y",
                self.coordformat)
        )
        radius = self.mpp.meter2px(
            int(self.element.find("annotation/radius").text), "r",
            self.coordformat)

        coords = []
        n = 100
        for i in range(0, n):
            coords.append([
                int(center.x + radius * math.cos(2*math.pi/n*i)),
                int(center.y + radius * math.sin(2*math.pi/n*i))
            ])
        return coords

    def get_pin_coords(self) -> List[int]:
        """Get the coordinate of Pin annotation"""
        coord = Coord(
            x=self.mpp.meter2px(
                int(self.element.find("annotation/x").text), "x",
                self.coordformat),
            y=self.mpp.meter2px(
                int(self.element.find("annotation/y").text), "y",
                self.coordformat)
        )
        return [
            [self.origin.x + coord.x, self.origin.y + coord.y]
        ]

    def get_polygon_coords(self) -> List[int]:
        """Get coordinates of rectangle, freehand, freehandline polygons.
        The coordinates are assumed to be closed.
        """
        coords = []
        for point in self.element.findall("annotation/pointlist/point"):
            x = self.mpp.meter2px(
                int(point.find("x").text), "x", self.coordformat)
            y = self.mpp.meter2px(
                int(point.find("y").text), "y", self.coordformat)
            coords.append([self.origin.x + x, self.origin.y + y])
        return coords


def get_origin(slide, mpp: MPP, coordformat: str):
    center = Coord(x=slide.width//2, y=slide.height//2)
    offset = Coord(
        x=int(getattr(slide, "hamamatsu.XOffsetFromSlideCentre")),
        y=int(getattr(slide, "hamamatsu.YOffsetFromSlideCentre"))
    )
    return Coord(
        x=center.x - mpp.meter2px(offset.x, "x", coordformat),
        y=center.y - mpp.meter2px(offset.y, "y", coordformat)
    )


def get_mpp(slide):
    mpp_x = float(getattr(slide, "openslide.mpp-x"))
    mpp_y = float(getattr(slide, "openslide.mpp-y"))
    mpp_r = (float(mpp_x) + float(mpp_y)) / 2
    return MPP(x=mpp_x, y=mpp_y, r=mpp_r)


class NDPViewAnnotation(BaseParser):
    """Annotation Parser for NDP View.
    https://www.hamamatsu.com/jp/ja/product/life-science-and-medical-systems/digital-slide-scanner/U12388-01.html

    Args:
        path (str): Path to the annotation file.
        slide (wsiprocess.Slide): Slide object.

    Attributes:
        path (str): Path to the annotation file.
        annotations (list): List of etree Elements.
        annotation_groups (list): List of etree Elements.
        classes (list): List of classes defined with ASAP.
        mask_coords (dict): Coordinates of the masks.
    """

    def __init__(self, path, slide):
        super().__init__(path)
        self.slide = slide
        self.tree = etree.parse(self.path)
        mpp = get_mpp(slide)
        origin = get_origin(
            slide, mpp, coordformat=self.tree.find("//coordformat").text)
        self.read_annotation(origin, mpp)

    def read_annotation(self, origin, mpp):
        for element in self.tree.xpath("/annotations/ndpviewstate"):
            annotation = Annotation(element, origin, mpp)
            self.classes.append(annotation.title)
            self.mask_coords[annotation.title].append(annotation.coords)



import json
import cv2
from pathlib import Path


def show_bounding_box(patch_path, result_path, save_as):
    img = cv2.imread(patch_path)
    with open(result_path, "r") as f:
        result = json.load(f)
    x, y = Path(patch_path).stem.split("_")
    for patch in result["result"]:
        if patch["x"] == int(x) and patch["y"] == int(y):
            for bbs in patch["bbs"]:
                x1 = bbs["x"]
                y1 = bbs["y"]
                x2 = bbs["x"] + bbs["w"]
                y2 = bbs["y"] + bbs["h"]
                color = (255, 255, 0)
                cv2.rectangle(img, (x1, y1), (x2, y2), color)
    cv2.imwrite(save_as, img)


def show_mask_on_patch(patch_path, mask_path, save_as):
    patch = cv2.imread(str(patch_path))
    mask = cv2.imread(str(mask_path)) * 255
    patch_with_mask = cv2.addWeighted(patch, 0.7, mask, 0.3, 0)
    cv2.imwrite(save_as, patch_with_mask)



import json
from pathlib import Path
from typing import Callable

import openslide
import torch
from torchvision import io, transforms
import pandas as pd

from wsiprocess import cli
from wsiprocess.pytorch import utils


class WSIDataset(torch.utils.data.Dataset):

    def __init__(
            self, path: str = False, command: str = None,
            foreground_fn: Callable = False):
        if path and Path(path).is_dir():
            # if wsiprocess is done, load from the result directory.
            # ex: slide=CMU-1.ndpi, save_to=/data/save_to
            #  -> self.path=/data/save_to/CMU-1
            self.path = Path(path)
        else:
            # if not done, extract the patch from the wsi_path.
            args = cli.Args(command)
            self.path = Path(args.save_to)/Path(args.wsi).stem
            if not (self.path/"results.json").exists():
                utils.main(command, foreground_fn=foreground_fn)
            else:
                print(f"skipped because already patched: {self.path}")

        with open(self.path/"results.json", "r") as f:
            self.patch_config = json.load(f)

        self.patch_extracted = not self.patch_config["no_patches"]
        if not self.patch_extracted:
            self.slide = openslide.OpenSlide(self.patch_config["slide"])
            self.read_patch = self.read_patch_from_wsi
            self.to_tensor = transforms.ToTensor()
        else:
            self.read_patch = self.read_patch_from_disk
            self.ext = self.patch_config["ext"]

        self.read_coords()

    def read_coords(self):
        self.coords = pd.read_csv(self.path/"coords.csv")

    def read_patch_from_wsi(self, **kwargs) -> torch.float32:
        x = kwargs["x"]
        y = kwargs["y"]
        w = kwargs["w"]
        h = kwargs["h"]
        patch = self.to_tensor(self.slide.read_region((x, y), 0, (w, h)))
        return patch

    def read_patch_from_disk(self, **kwargs) -> torch.float32:
        x = kwargs["x"]
        y = kwargs["y"]
        label = kwargs.get("label") or "foreground"
        path = str(self.path/"patches"/label/f"{x:06}_{y:06}.{self.ext}")
        patch = io.read_image(path, mode=io.ImageReadMode.RGB)/255
        return patch

    def read_patch(self):
        raise NotImplementedError()

    def __len__(self):
        return len(self.coords)

    def __getitem__(self, idx):
        coord = self.coords.iloc[idx].to_dict()
        patch = self.read_patch(**coord)
        return patch


class WSIsDataset(torch.utils.data.Dataset):

    def __init__(self, datasets: list):
        msg = "some wsi names might be duplicated"
        assert len(set([d.path for d in datasets])) == len(datasets), msg
        for d in datasets:
            d.coords["slide"] = d.patch_config["slide"]
        self.coords = pd.concat([d.coords for d in datasets])
        self.datasets = {d.patch_config["slide"]: d for d in datasets}

    def __len__(self):
        return len(self.coords)

    def __getitem__(self, idx):
        coord = self.coords.iloc[idx].to_dict()
        patch = self.datasets[coord["slide"]].read_patch(**coord)
        return patch



# -*- coding: utf-8 -*-
"""Convert wsiprocess style annotation data to COCO or VOC style.
"""
from .converters import (
    wsiprocess_to_coco,
    wsiprocess_to_voc,
    wsiprocess_to_yolo
)


class Converter:
    """Converter Class
    Args:

    Attributes:


    """
    def __init__(self, root, save_to, ratio_arg):
        self.params = {
            "root": root,
            "save_to": save_to,
            "ratio_arg": ratio_arg
        }

    def to_coco(self):
        converter = wsiprocess_to_coco.ToCOCOConverter(self.params)
        converter.convert()

    def to_voc(self):
        converter = wsiprocess_to_voc.ToVOCConverter(self.params)
        converter.convert()

    def to_yolo(self):
        converter = wsiprocess_to_yolo.ToYOLOConverter(self.params)
        converter.convert()



from .ASAP_parser import ASAPAnnotation
from .QuPath_parser import QuPathAnnotation
from .SlideRunner_parser import SlideRunnerAnnotation
from .WSIDissector_parser import WSIDissectorAnnotation
from .NDPView_parser import NDPViewAnnotation
from .parser_utils import BaseParser



import argparse
import os
from pathlib import Path
import json
import wsiprocess as wp


class Args:
    def __init__(self, command):
        self.build_args(command)
        self.fillattrs(keys=[
            "annotation", "rule", "export_thumbs", "on_annotation", "minmax",
            "crop_bbox", "extract_foreground"])

    def set_base_parser(self):
        self.base_parser = argparse.ArgumentParser(
            description="wsiprocess command line tool")

    def set_common_args(self, parser):
        """ Common Arguments """
        parser.add_argument(
            "-st", "--save_to", type=Path, default=".",
            help="Where to save the data.")
        parser.add_argument(
            "-pw", "--patch_width", type=int, default=256,
            help="Width of patches.")
        parser.add_argument(
            "-ph", "--patch_height", type=int, default=256,
            help="Height of patches.")
        parser.add_argument(
            "-ow", "--overlap_width", type=int, default=0,
            help="Width of the overlapped area of patches.")
        parser.add_argument(
            "-oh", "--overlap_height", type=int, default=0,
            help="Height of the overlapped area of patches")
        parser.add_argument(
            "-ox", "--offset_x", type=int, default=0,
            help="The offset pixel along the x-axis.")
        parser.add_argument(
            "-oy", "--offset_y", type=int, default=0,
            help="The offset pixel along the y-axis.")
        parser.add_argument(
            "-dw", "--dot_bbox_width", type=int, default=30,
            help="Width of bbox translated from dot annotation.")
        parser.add_argument(
            "-dh", "--dot_bbox_height", type=int,
            help="Height of bbox translated from dot annotation.")
        parser.add_argument(
            "-ex", "--ext", type=str, default="jpg",
            help="Extension of extracted patches")
        parser.add_argument(
            "-ma", "--magnification", type=int,
            help="Magnification of extracted patches")
        parser.add_argument(
            "-mw", "--max_workers", type=int, default=os.cpu_count()//2,
            help="The maximum number of workers to use in patching.")
        parser.add_argument(
            "-ss", "--start_sample", action="store_true",
            help="Generate samples at the start of the process.")
        parser.add_argument(
            "-fs", "--finished_sample", action="store_true",
            help="Generate samples at the end of the process.")
        parser.add_argument(
            "-np", "--no_patches", action="store_true",
            help="Patcher run without extracting patches.")
        parser.add_argument(
            "-ep", "--extract_patches", action="store_true",
            help="[Not Available]Extract the patches and save them as images.")
        parser.add_argument(
            "-ve", "--verbose", action="store_true",
            help="Show progress bar while patching.")
        parser.add_argument(
            "-dr", "--dryrun", action="store_true",
            help="Run patching only for first 100 patches.")

    def set_wsi_arg(self, parser):
        parser.add_argument(
            "wsi", type=str,
            help="Path to the target wsi.")

    def add_annotation_args(self, parser, slide_is_sparse=False):
        parser.add_argument(
            "annotation", type=str, default=False,
            help="Path to the annotation file.")
        parser.add_argument(
            "-ru", "--rule", type=Path,
            help="File to define the inclusion / exclusion relationship.")
        parser.add_argument(
            "-ef", "--extract_foreground", action="store_true",
            help="If set, wp extracts patches from foreground.")
        self.add_binarization_method(parser)
        self.add_on_foreground(parser, slide_is_sparse)
        self.add_on_annotation(parser, slide_is_sparse)

    def add_on_foreground(self, parser, slide_is_sparse=False):
        on_foreground_param = 0.0001 if slide_is_sparse else 0.01
        parser.add_argument(
            "-of", "--on_foreground", type=float, default=on_foreground_param,
            help="The ratio of overlapped area of a patch and the foreground.")

    def add_on_annotation(self, parser, slide_is_sparse=False):
        on_annotation_param = 0.0001 if slide_is_sparse else 0.01
        parser.add_argument(
            "-oa", "--on_annotation", type=json.loads, default=on_annotation_param,
            help="The ratio of overlapped area of a patch and the annotated.")

    def add_binarization_method(self, parser):
        parser.add_argument(
            "-mm", "--minmax", type=str, default=False,
            help="Get foreground mask as pixels from min to max. ie. 30-190")
        parser.add_argument(
            "-et", "--export_thumbs", action="store_true",
            help="Export thumbnails of masks.")

    def set_method_args(self):
        self.method_args = self.base_parser.add_subparsers(
            dest="method",
            help="Method to use.")
        self.method_args.required = True

    def set_evaluation_args(self):
        parser_eval = self.method_args.add_parser(
            "evaluation",
            help="Arguments for methods with no annotation data.")
        self.set_wsi_arg(parser_eval)
        self.add_on_foreground(parser_eval)
        self.add_binarization_method(parser_eval)
        self.set_common_args(parser_eval)

    def set_classification_args(self):
        """ Arguments for classification """
        parser_cls = self.method_args.add_parser(
            "classification",
            help="Arguments for classification tasks.")
        self.set_wsi_arg(parser_cls)
        self.add_annotation_args(parser_cls)
        self.set_common_args(parser_cls)

    def set_detection_args(self):
        """ Arguments for detection tasks. """
        parser_det = self.method_args.add_parser(
            "detection",
            help="Arguments for detection tasks.")
        self.set_wsi_arg(parser_det)

        parser_det.add_argument(
            "-vo", "--voc_style", action="store_true",
            help="Output as VOC style.")
        parser_det.add_argument(
            "-co", "--coco_style", action="store_true",
            help="Output as COCO style.")
        parser_det.add_argument(
            "-yo", "--yolo_style", action="store_true",
            help="Output as YOLO style.")
        parser_det.add_argument(
            "-ra", "--ratio", default="8:1:1",
            help="Ratio of the dataset size of train/validation/test phase.")
        parser_det.add_argument(
            "-cb", "--crop_bbox", default=False, action="store_true",
            help="Crop bounding boxes after patch extraction."
        )
        self.add_annotation_args(parser_det, slide_is_sparse=True)
        self.set_common_args(parser_det)

    def set_segmentation_args(self):
        """ Arguments for segmentation tasks. """
        parser_seg = self.method_args.add_parser(
            "segmentation",
            help="Arguments for segmentation tasksk."
        )
        self.set_wsi_arg(parser_seg)
        self.add_annotation_args(parser_seg)
        self.set_common_args(parser_seg)

    def build_args(self, command):
        """ Base Parser """
        self.set_base_parser()

        self.set_method_args()
        self.set_evaluation_args()
        self.set_classification_args()
        self.set_detection_args()
        self.set_segmentation_args()

        self.base_parser.parse_args(command, namespace=self)

    def fillattr(self, key: str):
        if not hasattr(self, key):
            setattr(self, key, False)

    def fillattrs(self, keys):
        for key in keys:
            self.fillattr(key)


def process_annotation(args, slide, rule):
    annotation = wp.annotation(args.annotation, slide=slide)
    annotation.dot_to_bbox(args.dot_bbox_width, args.dot_bbox_height)
    if args.minmax:
        min_, max_ = map(int, args.minmax.split("-"))
        annotation.make_masks(
            slide, rule, foreground_fn="minmax", min_=min_, max_=max_)
    else:
        annotation.make_masks(slide, rule, foreground_fn="otsu")

    return annotation


def main(command=None):
    args = Args(command)
    slide = wp.slide(args.wsi)
    rule = wp.rule(args.rule) if args.rule else False
    annotation = process_annotation(args, slide, rule)

    if args.export_thumbs:
        thumbs_dir = args.save_to/slide.filestem/"thumbs"
        thumbs_dir.mkdir(parents=True, exist_ok=True)
        annotation.export_thumb_masks(thumbs_dir)

    extract_classes = annotation.classes
    if not (args.extract_foreground and "foreground" in annotation.classes):
        extract_classes.remove("foreground")

    patcher = wp.patcher(
        slide, args.method,
        annotation=annotation,
        save_to=args.save_to,
        patch_width=args.patch_width,
        patch_height=args.patch_height,
        overlap_width=args.overlap_width,
        overlap_height=args.overlap_height,
        on_foreground=args.on_foreground,
        on_annotation=args.on_annotation,
        offset_x=args.offset_x,
        offset_y=args.offset_y,
        ext=args.ext,
        magnification=args.magnification,
        start_sample=args.start_sample,
        finished_sample=args.finished_sample,
        no_patches=args.no_patches,
        crop_bbox=args.crop_bbox,
        verbose=args.verbose,
        dryrun=args.dryrun)

    patcher.get_patch_parallel(
        extract_classes, max_workers=args.max_workers)

    if args.method == "detection":
        converter = wp.converter(
            args.save_to/slide.filestem,
            args.save_to/slide.filestem,
            args.ratio)
        if args.voc_style:
            converter.to_voc()
        if args.coco_style:
            converter.to_coco()
        if args.yolo_style:
            converter.to_yolo()

        if args.crop_bbox:
            patcher.get_mini_patch_parallel(annotation.classes)



# -*- coding: utf-8 -*-

import json

from .parser_utils import BaseParser


class WSIDissectorAnnotation(BaseParser):
    """Annotation parser for WSIDissector

    Args:
        path (str): Path to the annotation file.

    Attributes:
        path (str): Path to the annotation file.
        annotation (dict): Annotation data loaded as json file.
        filename (str): Name of the targeted whole slide image file.
        classes (list): List of the names of the classes.
        mask_coords (dict): Coordinates of the masks.
    """

    def __init__(self, path):
        super().__init__(path)

        with open(self.path, "r") as f:
            self.annotation = json.load(f)
        self.filename = self.annotation["slide"]
        self.classes = self.annotation["classes"]
        for cls in self.classes:
            self.mask_coords[cls] = []
        self.read_mask_coords()

    def read_mask_coords(self):
        """Parse the coordinates of masks.

        """
        annotations = self.annotation["result"]
        for annotation in annotations:
            cls = annotation["class"]
            contour = []
            if "points" in annotation:
                for point in annotation["points"]:
                    contour.append([point["x"], point["y"]])
            else:
                x = round(annotation["x"])
                y = round(annotation["y"])
                w = round(annotation["w"])
                h = round(annotation["h"])
                contour.append([x, y])
                contour.append([x+w, y])
                contour.append([x+w, y+h])
                contour.append([x, y+h])
            self.mask_coords[cls].append(contour)



# -*- coding: utf-8 -*-
"""Converter to VOC style.

wsiprocess_to_voc is convert helper which convert the wsiprocess output for
VOC object detection format.
Only available for "detection" method, but works when extracting patches and
after extraction.

"""
import argparse
import random
from lxml import etree
from pathlib import Path
from PIL import Image
import json
import shutil


class ToVOCConverter:
    """Converter class."""

    def __init__(self, params=False):
        """

        Args:
            params(dict): Dict contains root, save_to, ratio_arg

        'root' should be like below

        ./root
        ├── patches
        │   └── class_a
        │       ├── 001.png
        │       ├── 002.png

        │       └── 100.png
        └── results.json
        """
        if params:
            self.root = params["root"]
            self.save_to = params["save_to"]/"voc"
            self.ratio = params["ratio_arg"]
        else:
            self.getargs()
        self.makedirs()

    def convert(self):
        self.read_result_file()
        self.make_tree()
        self.move_to_test()

    def getargs(self):
        parser = argparse.ArgumentParser(description="wsiprocess_to_voc")
        parser.add_argument("root", type=Path)
        parser.add_argument("-s", "--save_to", default="./wp/voc", type=Path)
        parser.add_argument("-r", "--ratio", default="8:1:1")
        args = parser.parse_args()
        self.root = args.root
        self.save_to = args.save_to
        self.ratio = args.ratio

    def makedirs(self):
        self.mkdirs("VOC2007")
        self.mkdirs("VOC2012")
        Path(f"{self.save_to}/VOC2012/ImageSets/Main/trainval.txt").touch()

    def mkdirs(self, subdirectory):
        parent_dir = self.save_to/subdirectory
        (parent_dir/"Annotations").mkdir(exist_ok=True, parents=True)
        (parent_dir/"JPEGImages").mkdir(exist_ok=True, parents=True)
        (parent_dir/"ImageSets/Main").mkdir(exist_ok=True, parents=True)

    def read_result_file(self):
        with open(self.root/"results.json", "r") as f:
            self.result_wp = json.load(f)

    def make_tree(self):
        self.results = []
        for result in self.result_wp["result"]:
            x = result['x']
            y = result['y']
            tree = VOCTree(
                self.root, self.save_to/"VOC2007", result,
                self.result_wp["patch_width"], self.result_wp["patch_height"])
            tree.to_xml()
            cls = result["bbs"][0]["class"]
            src = f"{self.root}/patches/{cls}/{x:06}_{y:06}.jpg"
            dst_dir = f"{self.save_to}/VOC2007/JPEGImages"
            dst = f"{dst_dir}/{self.root.name}_{x:06}_{y:06}.jpg"
            shutil.copy(src, dst)
            self.results.append(f"{self.root.name}_{x:06}_{y:06}\n")

    def to_jpg(self, src, dst):
        img = Image.open(src).convert("RGB")
        imgname = f"{Path(src).parent.parent.parent.name}_{Path(src).stem}"
        img.save(f"{dst}/{imgname}.jpg", quality=95)

    def move_to_test(self):
        """
        If ratio has only two parameters like "8:2", trainval.txt and test.txt
        would be generated. If ratio has 3 params like "8:1:1", trainval.txt,
        val.txt and text.txt would be generated.
        """
        trainval_path = f"{self.save_to}/VOC2007/ImageSets/Main/trainval.txt"
        test_path = f"{self.save_to}/VOC2007/ImageSets/Main/test.txt"
        random.shuffle(self.results)

        test_is_available = len(self.ratio.split(":")) == 3
        if test_is_available:
            train, val, test = map(int, self.ratio.split(":"))
            ratio = {
                "train": train/(train+val+test),
                "val": val/(train+val+test),
                "test": test/(train+val+test)
            }
            train_count = int(len(self.results) * ratio["train"])
            val_count = int(len(self.results) * ratio["val"])
            val_path = f"{self.save_to}/VOC2007/ImageSets/Main/val.txt"

            with open(trainval_path, "a") as f:
                f.writelines(self.results[:train_count])
            with open(val_path, "a") as f:
                f.writelines(self.results[train_count:train_count+val_count])
            with open(test_path, "a") as f:
                f.writelines(self.results[train_count+val_count:])

        else:
            train, val = map(int, ratio.split(":"))
            ratio = {"train": train/(train+val), "val": val/(train+val)}
            train_count = int(len(self.results) * ratio["train"])

            with open(trainval_path, "a") as f:
                f.writelines(self.results[:train_count])
            with open(test_path, "a") as f:
                f.writelines(self.results[train_count:])


class VOCTree:

    def __init__(self, root, save_to, result, patch_width, patch_height):
        self.root = root
        self.save_to = save_to

        self.x = result["x"]
        self.y = result["y"]
        self.w = result["w"]
        self.h = result["h"]
        self.bbs = result["bbs"]
        self.patch_width = patch_width
        self.patch_height = patch_height
        self.imgname = f"{self.x:06}_{self.y:06}.png"
        self.out_name = f"{root.name}_{self.x:06}_{self.y:06}.xml"

        self.tree = etree.Element("annotation")
        self.main_branches = [
            "folder", "filename", "source", "owner", "size", "segmented"]
        self.sub_branches = {
            "source": ["database", "annotation", "image", "flickrid"],
            "owner": ["flickrid", "name"],
            "size": ["width", "height", "depth"]}

        self.make_template()
        self.set_size()
        self.add_objects()

    def make_template(self):
        self.add_main_branches()
        self.add_sub_branches()
        self.add_base_text()

    def add_main_branches(self):
        for branch in self.main_branches:
            etree.SubElement(self.tree, branch)

    def add_sub_branches(self):
        for main, sub_branches in self.sub_branches.items():
            main_branch = self.tree.find(main)
            for branch in sub_branches:
                etree.SubElement(main_branch, branch)

    def add_base_text(self):
        self.add_text("folder", "wsiprocess")
        self.add_text("filename", self.imgname)
        self.add_text("source/database", "original")
        self.add_text("/annotation/source/annotation", "PASCAL_style")

    def add_text(self, branch, text):
        target = self.tree.xpath(branch)[0]
        target.text = str(text)

    def set_size(self):
        # img_path = self.root/"patches"/"mitosis_figure"/self.imgname
        # img = Image.open(str(img_path))
        # w, h = img.size
        depth = 3  # len(img.getbands())
        self.add_text("/annotation/size/width", self.patch_width)
        self.add_text("/annotation/size/height", self.patch_height)
        self.add_text("/annotation/size/depth", depth)

    def add_objects(self):
        for bb in self.bbs:
            obj = etree.Element("object")
            name = etree.SubElement(obj, "name")
            name.text = bb["class"]
            etree.SubElement(obj, "pose")
            etree.SubElement(obj, "truncated")
            difficult = etree.SubElement(obj, "difficult")
            difficult.text = str(0)

            bndbox = etree.SubElement(obj, "bndbox")
            etree.SubElement(bndbox, "xmin")
            etree.SubElement(bndbox, "ymin")
            etree.SubElement(bndbox, "xmax")
            etree.SubElement(bndbox, "ymax")

            xmin = bb["x"]
            ymin = bb["y"]
            xmax = int(bb["x"] + int(bb["w"]))
            ymax = int(bb["y"] + int(bb["h"]))

            if xmin == xmax:
                if xmin == 0:
                    xmax += 1
                else:
                    xmin -= 1
            if ymin == ymax:
                if ymin == 0:
                    ymax += 1
                else:
                    ymin -= 1

            obj.xpath("/object/bndbox/xmin")[0].text = str(xmin)
            obj.xpath("/object/bndbox/ymin")[0].text = str(ymin)
            obj.xpath("/object/bndbox/xmax")[0].text = str(xmax)
            obj.xpath("/object/bndbox/ymax")[0].text = str(ymax)

            self.tree.insert(-1, obj)

    def to_xml(self):
        with open(f"{self.save_to}/Annotations/{self.out_name}", "wb") as f:
            out_tree = etree.ElementTree(self.tree)
            out_tree.write(f,
                           pretty_print=True,
                           xml_declaration=True,
                           encoding="UTF-8")


if __name__ == '__main__':
    converter = ToVOCConverter()
    converter.convert()



import json
from pathlib import Path

import torch
from torchvision import io

import wsiprocess as wp
from wsiprocess.cli import Args


class ClassificationDataset(torch.utils.data.Dataset):

    def __init__(self, dataset):
        self.dataset = dataset
        self.read_labels(self.dataset.coords)

    def read_labels(self, coords):
        normal_column = ["x", "y", "w", "h", "mask"]
        columns = list(set(coords.columns) - set(normal_column))
        self.labels = coords[columns]

    def __len__(self):
        return len(self.dataset.coords)

    def __getitem__(self, idx):
        coord = self.dataset.coords.iloc[idx]
        label = torch.tensor(self.labels.iloc[idx].tolist())
        return self.dataset.read_patch(**coord), label


class SegmentationDataset(torch.utils.data.Dataset):

    def __init__(self, dataset):
        if dataset.patch_config["no_patches"]:
            msg = "SegmentationDataset needs wsiprocess to have run without"
            msg += "no_patches because wsiprocess generates segmentation mask"
            msg += "from global view."
            raise NotImplementedError(msg)
        self.dataset = dataset
        self.path = Path(dataset.patch_config["save_to"]).parent
        self.read_masks(self.dataset.coords)

    def read_masks(self, coords):
        masks = coords.masks.str.replace("'", "\"")
        masks = masks.apply(lambda x: json.loads(x))

        self.masks = masks.apply(lambda x: x[0]["coords"])
        self.dataset.coords["label"] = masks.apply(lambda x: x[0]["class"])

    def __len__(self):
        return len(self.dataset.coords)

    def __getitem__(self, idx):
        coord = self.dataset.coords.iloc[idx]
        mask_path = str(self.path/self.masks.iloc[idx])
        mask = io.read_image(mask_path, mode=io.ImageReadMode.UNCHANGED)
        return self.dataset.read_patch(**coord), mask


def main(command, foreground_fn):
    args = Args(command)
    slide = wp.slide(args.wsi)
    rule = wp.rule(args.rule) if hasattr(args, "rule") and args.rule else False

    if args.method == "evaluation":
        annotation = wp.annotation()
    else:
        annotation = wp.annotation(args.annotation, slide=slide)
    annotation.dot_to_bbox(args.dot_bbox_width, args.dot_bbox_height)

    if foreground_fn:
        annotation.make_masks(slide, rule, foreground_fn=foreground_fn)
    elif hasattr(args, "minmax") and args.minmax:
        min_, max_ = map(int, args.minmax.split("-"))
        annotation.make_masks(
            slide, rule, foreground_fn="minmax", min_=min_, max_=max_)
    else:
        annotation.make_masks(slide, rule, foreground_fn="otsu")

    if hasattr(args, "extract_foreground"):
        if not (args.extract_foreground and "foreground" in annotation.classes):
            annotation.classes.remove("foreground")

    if hasattr(args, "export_thumbs") and args.export_thumbs:
        thumbs_dir = args.save_to/slide.filestem/"thumbs"
        if not thumbs_dir.exists():
            thumbs_dir.mkdir(parents=True)
        annotation.export_thumb_masks(thumbs_dir)

    if args.method == "evaluation":
        on_annotation = False
    else:
        on_annotation = args.on_annotation
    crop_bbox = args.crop_bbox if hasattr(args, "crop_bbox") else False

    patcher = wp.patcher(
        slide, args.method,
        annotation=annotation,
        save_to=args.save_to,
        patch_width=args.patch_width,
        patch_height=args.patch_height,
        overlap_width=args.overlap_width,
        overlap_height=args.overlap_height,
        on_foreground=args.on_foreground,
        on_annotation=on_annotation,
        offset_x=args.offset_x,
        offset_y=args.offset_y,
        ext=args.ext,
        magnification=args.magnification,
        start_sample=args.start_sample,
        finished_sample=args.finished_sample,
        no_patches=args.no_patches,
        crop_bbox=crop_bbox,
        verbose=args.verbose,
        dryrun=args.dryrun)
    patcher.get_patch_parallel(
        annotation.classes, max_workers=args.max_workers)

    if args.method == "detection":
        converter = wp.converter(
            args.save_to/slide.filestem,
            args.save_to/slide.filestem,
            args.ratio)
        if args.voc_style:
            converter.to_voc()
        if args.coco_style:
            converter.to_coco()
        if args.yolo_style:
            converter.to_yolo()

        if args.crop_bbox:
            patcher.get_mini_patch_parallel(annotation.classes)



# -*- coding: utf-8 -*-
"""Annotation object.

Annotation object is optional metadata for slide.
This object can handle ASAP or WSIViewer style annotation.
By adding annotationparser, you can process annotation data from other types of
annotation tools.


Example:
    Loading annotation data:: python

        import wsiprocess as wp
        annotation = wp.annotation("path_to_annotation_file.xml")

    Loading annotation data from image:: python

        import wsiprocess as wp
        annotation = wp.annotation("")
"""
from typing import Callable
import warnings
from pathlib import Path
from decimal import Decimal, ROUND_HALF_UP

import psutil
import cv2
import numpy as np
import wsiprocess.annotationparser as parsers
from .annotationparser.parser_utils import detect_type


class Annotation:

    def __init__(self, path=False, is_image=False, slide=False):
        """Initialize the anntation object.

        Args:
            path(str): Path to the anntation data.
                Text data made with "WSIDissector" or "ASAP", and Image data
                (non-pyramidical) are available.
            is_image(bool): Whether the image is image.

        Attributes:
            low_memory_consumption (bool): If true, annotaion object does not
                keep the processed masks on the ram, and read the area from
                disk when called get_patch.
        """
        self.path = path if path else ""
        self.slide = slide
        self.dot_bbox_width = self.dot_bbox_height = False
        self.is_image = is_image
        self.low_memory_consumption = False
        self.classes = []
        if not self.is_image:
            self.read_annotation()
        self.masks = {}
        self.contours = {}

    def __str__(self):
        return "wsiprocess.annotation.Annotation {}".format(self.path)

    def read_annotation(self, annotation_type=False):
        """Parse the annotation data.

        Args:
            annotation_type (str): If provided, pass the auto type detection.
        """
        if not annotation_type:
            annotation_type = detect_type(self.path)
        if annotation_type == "ASAP":
            parsed = parsers.ASAPAnnotation(self.path)
        elif annotation_type == "WSIDissector":
            parsed = parsers.WSIDissectorAnnotation(self.path)
        elif annotation_type == "SlideRunner":
            parsed = parsers.SlideRunnerAnnotation(self.path, self.slide)
        elif annotation_type == "QuPath":
            parsed = parsers.QuPathAnnotation(self.path)
        elif annotation_type == "GeoJson":
            parsed = parsers.GeoJsonAnnotation(self.path)
        elif annotation_type == "NDPView":
            parsed = parsers.NDPViewAnnotation(self.path, self.slide)
        elif annotation_type == "Empty":
            parsed = parsers.BaseParser(self.path)
        self.classes = parsed.classes
        self.mask_coords = parsed.mask_coords

    def dot_to_bbox(self, width=30, height=False):
        """Translate dot annotations to bounding boxes.

        If the len(self.mask_coords[cls][idx]) is 1, the annotation is a dot.
        And, the dot is the midpoint of the bounding box.

        Args:
            width (int): Width of the translated bounding box.
            height (int): Height of the translated bounding box. If not set,
                height is equal to width.
        """
        self.dot_bbox_width = width
        self.dot_bbox_height = width if not height else height

        for cls, coords in self.mask_coords.items():
            for idx, coord in enumerate(coords):
                if len(coord) == 1:
                    center_x = coord[0][0]
                    center_y = coord[0][1]
                    left = int(center_x - self.dot_bbox_width / 2)
                    top = int(center_y - self.dot_bbox_height / 2)
                    right = int(center_x + self.dot_bbox_width / 2)
                    bottom = int(center_y + self.dot_bbox_height / 2)
                    lefttop = [left, top]
                    righttop = [right, top]
                    leftbottom = [left, bottom]
                    rightbottom = [right, bottom]
                    self.mask_coords[cls][idx] = [
                        lefttop, righttop, rightbottom, leftbottom]

    def add_class(self, classes):
        for cls in classes:
            self.classes.append(cls)

    def from_image(self, mask, cls):
        """Load mask data from an image.

        Args:
            mask(numpy.ndarray): 2D mask image with background as 0, and
                foreground as 255.
            cls(str): Name of the class of the mask image.
        """
        assert len(mask.shape) == 2, "Mask image is not 2D."
        assert self.classes, "Classes are not set yet."
        self.masks[cls] = mask

    def make_masks(
            self, slide, rule=False, foreground_fn="otsu", size=5000,
            min_=30, max_=190):
        """Make masks from the slide and rule.

        Masks are for each class and foreground area.

        Args:
            slide (wsiprocess.slide.Slide): Slide object
            rule (:obj:`wsiprocess.rule.Rule`, optional): Rule object
            foreground_fn (str or callable, optional): This can be {otsu,
                minmax} or a user specified function.
            size (int, optional): Size of foreground mask.
            min (int, optional): Used if method is "minmax". Annotation object
                defines foreground as the pixels with the value between "min"
                and "max".
            max (int, optional): Used if method is "minmax". Annotation object
                defines foreground as the pixels with the value between "min"
                and "max".
        """
        if rule:
            self.check_classes(self.classes, rule.classes)
            self.classes = list(set(self.classes) & set(rule.classes))
        self.check_memory_consumption(slide.height, slide.width)
        self.set_scale(size, slide.height, slide.width)
        self.base_masks(size, slide.height, slide.width)
        self.main_masks(size, slide.height, slide.width)
        if foreground_fn:
            self.foreground_mask(
                slide, size, slide.height, slide.width, fn=foreground_fn,
                min_=min_, max_=max_)
            self.fix_mask_size()
        if rule:
            self.include_masks(rule)
            self.merge_include_coords(rule)
            self.exclude_masks(rule)
            # self.exclude_coords(rule)
        if not self.low_memory_consumption:
            self.resize_masks(slide.height, slide.width)

    def check_classes(self, annotation_class, rule_class):
        if set(annotation_class) != set(rule_class):
            msg = "classes in annotation and rule are different. "
            msg += f"got annotation: {annotation_class}, rule: {rule_class}"
            warnings.warn(msg)

    def check_memory_consumption(self, wsi_height, wsi_width):
        num_classes = len(self.classes) if self.classes else 1
        total_mask_size = num_classes*(wsi_height*wsi_width+120)
        mask_is_too_large = total_mask_size > psutil.virtual_memory().available
        if mask_is_too_large:
            msg = "Full size mask is too large for the RAM. "
            msg += "Running in low memory consumption mode."
            warnings.warn(msg)
            self.low_memory_consumption = True

    def set_scale(self, size, wsi_height, wsi_width):
        self.scale = self.get_scale(size, wsi_height, wsi_width)

    def get_scale(self, size, wsi_height, wsi_width):
        return size / max(wsi_height, wsi_width)

    def base_masks(self, size, wsi_height, wsi_width):
        """Make base masks.

        Args:
            size (int): The long side of masks.
            wsi_height (int): The height of wsi.
            wsi_width (int): The width of wsi.
        """
        scale = self.get_scale(size, wsi_height, wsi_width)
        mask_height = self._round(str(wsi_height * scale))
        mask_width = self._round(str(wsi_width * scale))

        for cls in self.classes:
            self.base_mask(cls, mask_height, mask_width)

    def _round(self, num):
        num = str(num)
        return int(Decimal(num).quantize(Decimal('0'), rounding=ROUND_HALF_UP))

    def base_mask(self, cls, mask_height, mask_width):
        """ Masks have same size of as the slide.

        Masks are canvases of 0s.

        Args:
            cls (str): Class name for each mask.
            mask_height (int): The height of base masks.
            mask_width (int): The width of base masks.
        """
        self.masks[cls] = np.zeros((mask_height, mask_width), dtype=np.uint8)

    def main_masks(self, size, wsi_height, wsi_width):
        """Main masks

        Write border lines following the rule and fill inside with 255.
        """

        scale = self.get_scale(size, wsi_height, wsi_width)
        for cls in self.classes:
            self.main_mask(cls, scale)

    def main_mask(self, cls, scale):
        contours = np.array(self.mask_coords[cls], dtype=object)
        for contour in contours:
            self.masks[cls] = cv2.drawContours(
                self.masks[cls],
                [np.int32(np.array(contour)*scale)], 0, 1,
                thickness=cv2.FILLED)

    def include_masks(self, rule):
        """Merge masks following the rule.

        Args:
            rule (wsiprocess.rule.Rule): Rule object.
        """
        self.masks_include = self.masks.copy()
        for cls in self.classes:
            if not hasattr(rule, cls):
                continue
            for include in getattr(rule, cls)["includes"]:
                if include not in self.masks:
                    continue
                self.masks_include[cls] = cv2.bitwise_or(
                    self.masks[cls], self.masks[include])
        self.masks = self.masks_include

    def merge_include_coords(self, rule):
        """Merge coordinations following the rule.

        Args:
            rule (wsiprocess.rule.Rule): Rule object.
        """
        for cls in self.classes:
            if not hasattr(rule, cls):
                continue
            for include in getattr(rule, cls)["includes"]:
                if include not in self.mask_coords.keys():
                    continue
                self.mask_coords[cls].extend(self.mask_coords[include])

    def exclude_masks(self, rule):
        """Exclude area from base mask with following the rule.

        Args:
            rule (wsiprocess.rule.Rule): Rule object.
        """
        self.masks_exclude = self.masks.copy()
        for cls in self.classes:
            if not hasattr(rule, cls):
                continue
            for exclude in getattr(rule, cls)["excludes"]:
                if exclude not in self.masks:
                    continue
                overlap_area = cv2.bitwise_and(
                    self.masks[cls], self.masks[exclude])
                self.masks_exclude[cls] = cv2.bitwise_xor(
                    self.masks_exclude[cls], overlap_area)
        self.masks = self.masks_exclude

    def exclude_coords(self, rule):
        """Exclude coordinations following the rule.

        Args:
            rule (wsiprocess.rule.Rule): Rule object.
        """
        for cls in self.classes:
            if not hasattr(rule, cls):
                continue
            base_set = [tuple(c) for c in self.mask_coords[cls]]
            for exclude in getattr(rule, cls)["excludes"]:
                exclude_set = [tuple(c) for c in self.mask_coords[exclude]]
                base_set -= exclude_set
            self.mask_coords[cls] = [list(c) for c in base_set]

    def foreground_mask(
            self, slide, size=5000, wsi_height=False, wsi_width=False,
            fn="otsu", min_=30, max_=190):
        """Make foreground mask.

        With otsu thresholding, make simple foreground mask.

        Args:
            slide (wsiprocess.slide.Slide): Slide object.
            size (int, or function, optional): Size of foreground mask on
                calculating with the Otsu Thresholding.
            wsi_width (int or bool): Width of the wsi.
            wsi_height (int or bool): Height of the wsi.
            fn (str or function, optional): Binarization method. As default,
                calculates with Otsu Thresholding.
            min (int, optional): Used if method is "minmax". Annotation object
                defines foreground as the pixels with the value between "min"
                and "max".
            max (int, optional): Used if method is "minmax". Annotation object
                defines foreground as the pixels with the value between "min"
                and "max".
        """
        if "foreground" in self.classes:
            print("foreground is already calculated.")
            return

        if wsi_width and wsi_height:
            scale = self.get_scale(size, wsi_height, wsi_width)
            thumb_height = self._round(str(wsi_height * scale))
            thumb_width = self._round(str(wsi_width * scale))
            thumb_size = (thumb_width, thumb_height)
        else:
            thumb_size = size

        thumb = np.asarray(slide.get_thumbnail(thumb_size))
        thumb_gray = cv2.cvtColor(thumb, cv2.COLOR_RGB2GRAY)
        if isinstance(fn, str):
            if fn == "minmax":
                mask = self._minmax_mask(thumb_gray, min_, max_)
            elif fn == "otsu":
                mask = self._otsu_method_mask(thumb_gray)
            else:
                raise NotImplementedError(
                    "{} is not implemented for making masks.".format(fn))
        elif isinstance(fn, Callable):
            mask = fn(thumb_gray)
        else:
            raise NotImplementedError(
                "{} is not implemented for making masks.".format(fn))
        self.masks["foreground"] = mask
        self.classes.append("foreground")

    def fix_mask_size(self):
        if "foreground" not in self.masks:
            warnings.warn("foreground mask is not set yet.")
            return

        classes = self.classes.copy()
        if len(classes) == 1 and classes[0] == "foreground":
            return
        classes.remove("foreground")
        height, width = self.masks[classes[0]].shape
        if self.masks["foreground"].shape != (height, width):
            self.masks["foreground"] = cv2.resize(self.masks["foreground"], (width, height))

    def resize_masks(self, wsi_height, wsi_width):
        """Resize the masks as the same size as the slide

        Args:
            slide (wsiprocess.slide.Slide): Slide object.
        """
        for cls in self.classes:
            self.resize_mask(wsi_height, wsi_width, cls)

    def resize_mask(self, wsi_height, wsi_width, cls):
        """Resize a mask as the same size as the slide"""

        self.masks[cls] = cv2.resize(
            self.masks[cls], (wsi_width, wsi_height))

    def _otsu_method_mask(self, thumb_gray):
        """Make mask of foreground with Otsu's method.

        Foreground as 1, background as 0.

        Args:
            thumb_gray (numpy.ndarray): Mask image from 0 to 255.

        Returns:
            mask (numpy.ndarray): Binary mask image.
        """
        _, mask = cv2.threshold(
            thumb_gray, 0, 1, cv2.THRESH_BINARY_INV+cv2.THRESH_OTSU)
        return mask

    def _minmax_mask(self, thumb_gray, min_, max_):
        """Make mask of foreground from min and max value.

        Pixels with value between min_ and max_ is converted to 1 as
        foreground, and value less than min_ and more than max_ is converted to
        0 as background.

        Args:
            thumb_gray (numpy.ndarray): Mask image from 0 to 255.
            min_ (int): Minimum value of thumb_gray to convert to 1.
            max_ (int): Maximum value of thumb_gray to convert to 1.

        Returns:
            mask (numpy.ndarray): Binary mask image.
        """
        mask = ((min_ <= thumb_gray) & (thumb_gray <= max_)).astype(np.uint8)
        return mask

    def export_thumb_masks(self, save_to=".", size=512):
        """Export thumbnail of masks.

        For prior check, export thumbnails of masks.

        Args:
            save_to (str): Parent directory to save the thumbnails.
            size (int): Length of the long side of thumbnail.
        """
        for cls in self.masks.keys():
            self.export_thumb_mask(cls, save_to, size)

    def export_thumb_mask(self, cls, save_to=".", size=512):
        """Export a thumbnail of one of the masks.

        For prior check, export one thumbnail of one of the masks.

        Args:
            cls (str): Class name for each mask.
            save_to (str, optional): Parent directory to save the thumbnails.
            size (int, optional): Length of the long side of thumbnail.
        """
        mask = self.masks[cls]
        height, width = mask.shape
        scale = self.get_scale(size, height, width)
        mask_resized = cv2.resize(mask, dsize=None, fx=scale, fy=scale)
        mask_scaled = mask_resized * 255
        cv2.imwrite(str(Path(save_to)/"{}_thumb.png".format(cls)), mask_scaled)

    def export_masks(self, save_to):
        """Export binary mask images.

        For later computing such as segmentation, export the mask images.
        Exported masks have 0 or 1 binary data.

        Args:
            save_to (str): Parent directory to save the thumbnails.
        """
        for cls in self.masks.keys():
            self.export_mask(save_to, cls)

    def export_mask(self, save_to, cls):
        """Export one binary mask image.

        Export mask image with 0 or 1 binaries.

        Args:
            save_to (str): Parent directory to save the thumbnails.
            cls (str): Class name for each mask.
        """
        cv2.imwrite(
            str(Path(save_to)/"{}.png".format(cls)),
            self.masks[cls], (cv2.IMWRITE_PXM_BINARY, 1))

    def get_patch_mask(self, cls, x, y, w, h):
        if self.low_memory_consumption:
            x_ = int(x * self.scale)
            y_ = int(y * self.scale)
            w_ = int(w * self.scale)
            h_ = int(h * self.scale)
            patch_mask = self.masks[cls][y_:y_+h_, x_:x_+w_]
            return cv2.resize(patch_mask, dsize=(w, h))

        else:
            return self.masks[cls][y:y+h, x:x+w]



# -*- coding: utf-8 -*-
from lxml import etree
import numpy as np

from .parser_utils import BaseParser


class ASAPAnnotation(BaseParser):
    """Annotation Parser for ASAP.

    Args:
        path (str): Path to the annotation file.

    Attributes:
        path (str): Path to the annotation file.
        annotations (list): List of etree Elements.
        annotation_groups (list): List of etree Elements.
        classes (list): List of classes defined with ASAP.
        mask_coords (dict): Coordinates of the masks.
    """

    def __init__(self, path):
        super().__init__(path)

        tree = etree.parse(self.path)
        self.annotations = tree.xpath(
            "/ASAP_Annotations/Annotations/Annotation")
        self.annotation_groups = tree.xpath(
            "/ASAP_Annotations/AnnotationGroups/Group")
        self.classes = [
            group.attrib["Name"]
            for group in self.annotation_groups
        ]
        for cls in self.classes:
            self.mask_coords[cls] = []
        self.read_mask_coords()

    def read_mask_coords(self):
        """Parse coordinates of of the masks of all classes.
        """
        for cls in self.classes:
            self.read_mask_coord(cls)

    def read_mask_coord(self, cls):
        """Parse the coordinates of the mask.

        Args:
            cls (str): Target annotation class name.
        """
        for annotation in self.annotations:
            if annotation.attrib["PartOfGroup"] == cls:
                contour = []
                for coord in annotation.xpath("Coordinates/Coordinate"):
                    x = np.float64(coord.attrib["X"])
                    y = np.float64(coord.attrib["Y"])
                    contour.append([round(float(x)), round(float(y))])
                self.mask_coords[cls].append(contour)



# -*- coding: utf-8 -*-
"""Custom errors for wsiprocess.
"""


class WsiProcessError(Exception):
    """Root error class.

    Args:
        message (str): Message to show in the stdout.
    """
    def __init__(self, message):
        self.message = message


class SlideLoadError(WsiProcessError):
    """Error on loading slides.

    Args:
        message (str): Message to show in the stdout.
    """
    def __init__(self, message):
        super().__init__(message)


class AnnotationLabelError(WsiProcessError):
    """Error of annotations

    Args:
        message (str): Message to show in the stdout.
    """
    def __init__(self, message):
        super().__init__(message)


class MissCombinationError(WsiProcessError):
    """Error of the combination of the method and the anntoation file.

    Args
        message (str): Message to show in the stdout.
    """
    def __init__(self, message):
        super().__init__(message)


class PatchSizeTooSmallError(WsiProcessError):
    """Error of the size of patches.

    This should be warning?

    Args:
        message (str): Message to show in the stdout.
    """
    def __init__(self, message):
        super().__init__(message)


class SizeError(WsiProcessError):
    """Error of sizes.

    The slide size is larger than the patches, and the patch size is larger
    than the overlap size.

    Args:
        message (str): Message to show in the stdout.
    """
    def __init__(self, message):
        super().__init__(message)


class OnParamError(WsiProcessError):
    """Error of on_annotation.

    on_annotation must be more than 0 and up to 1.

    Args:
        message (str): Message to show in the stdout.
    """
    def __init__(self, message):
        super().__init__(message)



# -*- coding: utf-8 -*-
"""Verification script runs before the patcher works.
Verify class works for verification of the output directory, annotation files,
rule files, etc. Mainly runs for cli.
"""

import warnings
from pathlib import Path
from .error import SizeError, OnParamError


class Verify:
    """Verification class.

    Args:
        save_to (str): The root of the output directory.
        filestem (str): The name of the output directory.
        method (str): Method name to run. One of {"none", "classification",
            "detection", "segmentation}
        start_sample (bool): Whether to save sample patches on Patcher start.
        finished_sample (bool): Whether to save sample patches on Patcher
            finish.
        extract_patches (bool): [Deleted]Whether to save patches when Patcher
            runs.
        no_patches (bool): Whether to save patches when Patcher runs.

    Attributes:
        save_to (str): The root of the output directory.
        filestem (str): The name of the output directory.
        method (str): Method name to run. One of {"none", "classification",
            "detection", "segmentation}
        start_sample (bool): Whether to save sample patches on Patcher start.
        finished_sample (bool): Whether to save sample patches on Patcher
            finish.
        extract_patches (bool): [Deleted]Whether to save patches when Patcher
            runs.
        no_patches (bool): Whether to save patches when Patcher runs.
    """

    def __init__(
            self, save_to, filestem, method, start_sample,
            finished_sample, no_patches, crop_bbox):
        self.save_to = save_to
        self.filestem = filestem
        self.method = method
        self.start_sample = start_sample
        self.finished_sample = finished_sample
        self.no_patches = no_patches
        self.crop_bbox = crop_bbox

    def make_dirs(self):
        """Ensure the output directories exists for each tasks.
        """
        base_dir = Path(self.save_to)/self.filestem
        if base_dir.exists():
            warnings.warn(
                "saving results to {}, but it already exists.".format(base_dir)
            )
        self.make_dir(base_dir)
        if self.method == "evaluation" and not self.no_patches:
            self.make_dir(base_dir/"patches"/"foreground")
        if self.method == "segmentation":
            self.make_dir(base_dir/"masks")
        if self.start_sample:
            self.make_dir(base_dir/"start_sample")
        if self.finished_sample:
            self.make_dir(base_dir/"finished_sample")
        if self.crop_bbox:
            self.make_dir(base_dir/"mini_patches")

    @staticmethod
    def make_dir(path):
        """Make output directory.
        """
        if not Path(path).exists():
            Path(path).mkdir(parents=True)

    def magnification(self, slide, magnification):
        """Check if the slide has data for the magnification the user specified.

        Args:
            slide (wp.slide.Slide): Slide object to check.
            magnification (int): Target magnification value which has to be
                smaller than the magnification of the slide.
        """
        basemsg = "Magnification for this slide has to be smaller than"
        msg = "{} {}".format(basemsg, slide.slide.magnification)
        assert slide.slide.magnification < magnification, msg

    def sizes(
            self, wsi_width, wsi_height, offset_x, offset_y,
            patch_width, patch_height, overlap_width, overlap_height,
            dot_bbox_width=False, dot_bbox_height=False):
        """Verify the sizes of the slide, the patch and the overlap area.

        Raises:
            wsiprocess.error.SizeError: If the sizes are invalid.
        """
        if wsi_width < patch_width or wsi_height < patch_height:
            raise SizeError("WSI have to be larger than the patches.")
        if wsi_width < offset_x or wsi_height < offset_y:
            raise SizeError("Offset must be less than wsi size.")
        if patch_width <= overlap_width or patch_height <= overlap_height:
            raise SizeError("Patches have to be larger than the overlap size.")
        if patch_width < 0 or patch_height < 0:
            raise SizeError("Patches has to be larger than 1.")
        if dot_bbox_width is not False and dot_bbox_width > patch_width:
            raise SizeError("Translated bbox is larger than patch width.")
        if dot_bbox_height is not False and dot_bbox_height > patch_height:
            raise SizeError("Translated bbox is larger than patch height.")

    def on_params(self, on_annotation, on_foreground):
        """Verify the ratio of on_annotation.

        Args:
            on_annotation (float): Overlap ratio of patches and annotations.
            on_foreground (float): Overlap ratio of patches and foreground area

        Raises:
            wsiprocess.error.SizeError: If the sizes are invalid.
        """
        if on_annotation is not False:
            if isinstance(on_annotation, (float, int)):
                if on_annotation <= 0 or on_annotation > 1:
                    raise OnParamError(
                        "on_annotation is between 0 and 1 excluding 0.")
            elif isinstance(on_annotation, dict):
                for v in on_annotation.values():
                    if v <= 0 or v > 1:
                        raise OnParamError(
                            "on_annotation is between 0 and 1 excluding 0.")

        if on_foreground is not False:
            if on_foreground <= 0 or on_foreground > 1:
                raise OnParamError(
                    "on_foreground is between 0 and 1 excluding 0.")



# -*- coding: utf-8 -*-
from pathlib import Path
from collections import defaultdict
from lxml import etree
import json
import sqlite3


def detect_type(path):
    """Detect the type of input file.

    Returns:
        file_type (str): One of {"ASAP", "WSIDissector", "SlideRunner",
            "WSIDissector", "Empty"}
    """
    path = Path(path)
    if path.suffix == ".xml":
        tree = etree.parse(str(path))
        root = tree.getroot()
        if root.tag == "ASAP_Annotations":
            return "ASAP"
    elif path.suffix == ".json":
        with open(path, "r") as f:
            data = json.load(f)
        if "annotationTool" in data:
            if data["annotationTool"] == "WSIDissector":
                return "WSIDissector"
        elif data[0].keys() >= {"type", "id", "geometry", "properties"}:
            return "QuPath"
        elif data.get("type") == "FeatureCollection":
            return "GeoJson"
    elif path.suffix == ".sqlite":
        try:
            con = sqlite3.connect(path)
            cursor = con.cursor()
            cursor.execute("select * from sqlite_sequence")
            assumed = set([
                "Persons", "Classes", "Slides", "Annotations",
                "Annotations_coordinates", "Annotations_label", "Log"])
            if assumed == set([x[0] for x in cursor.fetchall()]):
                file_type = "SlideRunner"
        except Exception as e:
            print(e)
    elif path.suffix == ".ndpa":
        file_type = "NDPView"
    else:
        file_type = "Empty"
    return file_type


class BaseParser:
    """Base class for Parsers"""

    def __init__(self, path):
        """
        Args:
            path (str): Path to the annotation file.
        Attributes:
            path (str): Path to the annotation file.
        """
        self.path = path
        assert Path(self.path).exists(), "This annotation file does not exist."

        self.classes = []
        self.mask_coords = defaultdict(list)



# -*- coding: utf-8 -*-
"""Patcher object to extract patches from whole slide images.
"""

import warnings
import random
from itertools import product
import json
import os
from concurrent.futures import ThreadPoolExecutor

from tqdm import tqdm
import numpy as np
import cv2
from pathlib import Path
import pandas as pd

from .verify import Verify


class Patcher:
    """Patcher object.

    Args:
        slide (wsiprocess.slide.Slide): Slide object.
        method (str): Method name to run. One of {"evaluation",
            "classification", "detection", "segmentation}. Characters are
            converted to lowercase.
        annotation (wsiprocess.annotation.Annotation, optional):
            Annotation object.
        save_to (str, optional): The root of the output directory.
        patch_width (int, optional): The width of the output patches.
        patch_height (int, optional): The height of the output patches.
        overlap_width (int, optional): The width of the overlap areas of
            patches.
        overlap_height (int, optional): The height of the overlap areas of
            patches.
        offset_x (int, optional): The offset pixels along the x-axis.
        offset_y (int, optional): The offset pixels along the y-axis.
        on_foreground (float, optional): Ratio of overlap area between patches
            and foreground area.
        on_annotation (float or dict): Ratio of overlap area between patches
            and annotation. dict is also available ex: {"label":value}.
        magnification (int, optional): Magnification of output patches.
        ext (str, optional): Extension of extracted patches.
        start_sample (bool, optional): Whether to save sample patches on
            Patcher starting.
        finished_sample (bool, optional): Whether to save sample patches on
            Patcher finished its work.
        extract_patches (bool, optional): This is deprecated because unless
            "no_patches" is set, Patcher extracts patches.
        no_patches (bool, optional): If set, Patcher runs without extracting
            patches and saves them to disk.
        verbose (bool, optional): If set, a progress bar appears when patching.
        dryrun (bool, optional): Only run patching for first 100 patches.

    Attributes:
        slide (wsiprocess.slide.Slide): Slide object.
        wsi_width (int): Width of the slide.
        wsi_height (int): Height of the slide.
        filepath (str): Path to the whole slide image.
        filestem (str): Stem of the file name.
        method (str): Method name to run. One of {"evaluation",
            "classification", "detection", "segmentation}
        annotation (wsiprocess.annotation.Annotation):
            Annotation object.
        masks (dict): Masks to show the location of classes.
        classes (list): Classes to extract.
        save_to (str): The root of the output directory.
        p_width (int): The width of the cropped patches. It magnification is
            not set, it is same as output patches.
        p_height (int): The height of the cropped patches. It magnification is
            not set, it is same as output patches.
        p_area (int): The area of single patch.
        o_width (int): The width of the overlap areas of patches.
        o_height (int): The height of the overlap areas of patches.
        offset_x (int): The The offset pixels along the x-axis.
        offset_y (int): The The offset pixels along the y-axis.
        on_foreground (float): Ratio of overlap area between patches and
            foreground area.
        on_annotation (float or dict): Ratio of overlap area between patches
            and annotation. dict is also available ex: {"label":value}.
        magnification (int): Magnification of output patches.
        p_scale (float): Ratio of p_width or p_height to output patch size.
        ext (str): Extension of extracted patches.
        start_sample (bool): Whether to save sample patches on Patcher start.
        finished_sample (bool): Whether to save sample patches on Patcher
            finish.
        extract_patches (bool): Whether to save patches when Patcher runs.
        no_patches (bool): Whether to save patches when Patcher runs.
        verbose (bool, optional): If set, a progress bar appears when patching.
        dryrun (bool, optional): Only run patching for first 100 patches.

        x_lefttop (list): Offsets of patches to the x-axis direction except for
            the right edge.
        y_lefttop (list): Offsets of patches to the y-axis direction except for
            the bottom edge.
        iterator (list):  Offset coordinates of patches.
        last_x (int): X-axis offset of the right edge patch.
        last_y (int): Y-axis offset of the right edge patch.

        result (dict): Temporary storage for the computed result of patches.
    """

    def __init__(
            self, slide, method, annotation=False, save_to=".",
            patch_width=256, patch_height=256, overlap_width=0,
            overlap_height=0, offset_x=0, offset_y=0, on_foreground=0.5,
            on_annotation=0.5, ext="jpg", magnification=False,
            start_sample=False, finished_sample=False, no_patches=False,
            crop_bbox=False, verbose=False, dryrun=False):
        self.verify = Verify(
            save_to, slide.filestem, method, start_sample, finished_sample,
            no_patches, crop_bbox)
        self.verify.sizes(
            slide.width, slide.height, offset_x, offset_y,
            patch_width, patch_height, overlap_width, overlap_height,
            annotation.dot_bbox_width, annotation.dot_bbox_height)
        self.verify.on_params(on_annotation, on_foreground)
        self.verify.make_dirs()

        self.slide = slide
        self.filepath = slide.filename
        self.filestem = slide.filestem
        self.method = method.lower()
        self.wsi_width = slide.width
        self.wsi_height = slide.height
        self.set_magnification(slide, magnification)
        self.p_width = int(patch_width*self.p_scale)
        self.p_height = int(patch_height*self.p_scale)
        self.p_area = self.p_width * self.p_height
        self.o_width = int(overlap_width)
        self.o_height = int(overlap_height)
        self.offset_x = int(offset_x)
        self.offset_y = int(offset_y)
        self.dot_bbox_width = annotation.dot_bbox_width
        self.dot_bbox_height = annotation.dot_bbox_height
        self.x_lefttop = [i for i in range(
            self.offset_x,
            self.wsi_width,
            patch_width - overlap_width)][:-1]
        self.y_lefttop = [i for i in range(
            self.offset_y,
            self.wsi_height,
            patch_height - overlap_height)][:-1]

        self.last_x = self.slide.width - patch_width
        self.last_y = self.slide.height - patch_height

        self.dryrun = dryrun
        self.get_iterator(dryrun)

        self.ext = ext

        self.start_sample = start_sample
        self.finished_sample = finished_sample
        self.no_patches = no_patches

        self.verbose = verbose

        self.on_foreground = on_foreground
        self.annotation = annotation
        if annotation:
            self.masks = annotation.masks
            self.classes = annotation.classes
            if isinstance(on_annotation, (float, int)):
                # if float, apply same threshold value for all classes
                self.on_annotation = {cl: on_annotation for cl in self.classes}
            else:
                self.on_annotation = on_annotation
        else:
            self.masks = False
            self.classes = []
            self.on_annotation = False

        self.save_to = save_to

        self.result = {"result": []}

    def __str__(self):
        return "wsiprocess.patcher.Patcher {}".format(self.slide.path)

    def get_iterator(self, dryrun=False):
        self.iterator = list(product(self.x_lefttop, self.y_lefttop))
        bottomedge = [(x, self.last_y) for x in self.x_lefttop]
        rightedge = [(self.last_x, y) for y in self.y_lefttop]
        rightbottom = [(self.last_x, self.last_y)]
        self.iterator += bottomedge
        self.iterator += rightedge
        self.iterator += rightbottom

        if dryrun:
            self.iterator = self.iterator[:100]

    def set_magnification(self, slide, magnification):
        self.magnification = magnification
        if self.magnification:
            if slide.magnification is None:
                raise KeyError(
                    "{} has no magnification property".format(slide.path))
            if slide.magnification % self.magnification != 0:
                msg = "magnification={} is not a factor of \
                    slide.magnification={}"
                warnings.warn(
                    msg.format(self.magnification, slide.magnification))
            self.p_scale = slide.magnification / self.magnification
        else:
            self.p_scale = 1

    def save_patch_result(self, x, y, cls):
        """Save the extracted patch data to result

        Args:
            x (int): X-axis offset of patch.
            y (int): Y-axis offset of patch.
            cls (str): Class of the patch or the bounding box or the segmented
                area.
        """
        if self.method == "evaluation":
            self.result["result"].append({"x": x,
                                          "y": y,
                                          "w": self.p_width,
                                          "h": self.p_height})

        elif self.method == "classification":
            self.result["result"].append({"x": x,
                                          "y": y,
                                          "w": self.p_width,
                                          "h": self.p_height,
                                          "class": cls})

        elif self.method == "detection":
            bbs = []
            for cls in self.classes:
                for bb in self.find_bbs(x, y, cls):
                    bbs.append({"x": bb["x"],
                                "y": bb["y"],
                                "w": bb["w"],
                                "h": bb["h"],
                                "class": bb["class"]})
            if bbs:
                self.result["result"].append({"x": x,
                                              "y": y,
                                              "w": self.p_width,
                                              "h": self.p_height,
                                              "bbs": bbs})

        elif self.method == "segmentation":
            masks = []
            for cls in self.classes:
                for mask in self.find_masks(x, y, cls):
                    masks.append({"coords": mask["coords"],
                                  "class": mask["class"]})
            self.result["result"].append({"x": x,
                                          "y": y,
                                          "w": self.p_width,
                                          "h": self.p_height,
                                          "masks": masks})

        else:
            raise NotImplementedError

    def find_bbs(self, x, y, cls):
        """Find bounding boxes which are on the patch.

        Bounding boxes with one of its corners on the patch is on the patch.
            ex : annotation.mask_coords["benign"][0]
             = [small_x, small_y, large_x, large_y]
             = [bbleft, bbtop, bbright, bbbottom]

        Args:
            x (int): X-axis offset of patch.
            y (int): Y-axis offset of patch.
            cls (str): Class of the patch or the bounding box or the segmented
                area.

        """
        if not self.patch_on_annotation(cls, x, y):
            return []
        else:
            # Find bounding boxes which are on the patch
            if cls == "foreground":
                return []
            coords = self.annotation.mask_coords[cls]
            for idx, coord in enumerate(coords):
                if len(coord) != 4:
                    coords[idx] = self.to_bb(coord)
            coords = np.array(coords)

            idx_of_bb_on_patch = self.corner_on_patch(coords, x, y)
            idx_of_bb_on_patch += self.side_on_patch(coords, x, y)
            idx_of_bb_on_patch += self.annotation_cover_patch(coords, x, y)

            idx_of_bb_on_patch = list(set(idx_of_bb_on_patch))

            bbs_raw = coords[idx_of_bb_on_patch]
            bbs = []
            for bb_raw in bbs_raw:
                x1 = bb_raw[:, 0].min()
                y1 = bb_raw[:, 1].min()
                x2 = bb_raw[:, 0].max()
                y2 = bb_raw[:, 1].max()
                bbx = int(max(x1 - x, 0))
                bby = int(max(y1 - y, 0))
                bbw = int(min(x2 - x1 + bbx, self.p_width)) - bbx
                bbh = int(min(y2 - y1 + bby, self.p_height)) - bby
                bb = {"x": bbx,
                      "y": bby,
                      "w": bbw,
                      "h": bbh,
                      "class": cls}
                bbs.append(bb)
            return bbs

    def corner_on_patch(self, coords, x, y):
        """Check if at least one of the corners is on the patch.

        Args:
            coords (np.array) : Coordinations of annotations.
            px (int) : X coordinate of left top corner of the patch.
            py (int) : Y coordinate of left top corner of the patch.

        Returns:
            idx_of_bb_on_patch (list): List of np.int64s which are the indices
                of bounding boxes on the patch.
        """
        bblefts = np.min(coords, axis=1)[:, 0]
        bbtops = np.min(coords, axis=1)[:, 1]
        bbrights = np.max(coords, axis=1)[:, 0]
        bbbottoms = np.max(coords, axis=1)[:, 1]

        patch_left = x
        patch_right = x + self.p_width
        patch_top = y
        patch_bottom = y + self.p_height

        left_on_patch = (patch_left <= bblefts) & (bblefts <= patch_right)
        right_on_patch = (patch_left <= bbrights) & (bbrights <= patch_right)
        top_on_patch = (patch_top <= bbtops) & (bbtops <= patch_bottom)
        bottom_on_patch = (patch_top <= bbbottoms) & (
            bbbottoms <= patch_bottom)

        topleft_on_patch = left_on_patch & top_on_patch
        topright_on_patch = right_on_patch & top_on_patch
        bottomleft_on_patch = left_on_patch & bottom_on_patch
        bottomright_on_patch = right_on_patch & bottom_on_patch

        idx_of_bb_on_patch = topleft_on_patch | topright_on_patch |\
            bottomleft_on_patch | bottomright_on_patch
        idx_of_bb_on_patch = np.where(idx_of_bb_on_patch)[0]
        return list(idx_of_bb_on_patch)

    def side_on_patch(self, coords, x, y):
        """Check if at least one of the side is on the patch.

        Args:
            coords (np.array) : Coordinations of annotations.
            px (int) : X coordinate of left top corner of the patch.
            py (int) : Y coordinate of left top corner of the patch.

        Returns:
            idx_of_bb_on_patch (list): List of np.int64s which are the indices
                of bounding boxes on the patch.
        """
        bblefts = np.min(coords, axis=1)[:, 0]
        bbtops = np.min(coords, axis=1)[:, 1]
        bbrights = np.max(coords, axis=1)[:, 0]
        bbbottoms = np.max(coords, axis=1)[:, 1]

        patch_left = x
        patch_right = x + self.p_width
        patch_top = y
        patch_bottom = y + self.p_height

        left_on_patch = (patch_left <= bblefts) & (bblefts <= patch_right)
        right_on_patch = (patch_left <= bbrights) & (bbrights <= patch_right)
        top_on_patch = (patch_top <= bbtops) & (bbtops <= patch_bottom)
        bottom_on_patch = (patch_top <= bbbottoms) & (
            bbbottoms <= patch_bottom)

        # One side is on the patch
        leftside_on_patch = (left_on_patch) &\
                            (bbtops <= patch_top) &\
                            (patch_bottom <= bbbottoms)
        rightside_on_patch = (right_on_patch) &\
                             (bbtops <= patch_top) &\
                             (patch_bottom <= bbbottoms)
        topside_on_patch = (bblefts <= patch_left) &\
                           (patch_right <= bbrights) &\
                           (top_on_patch)
        bottomside_on_patch = (bblefts <= patch_left) &\
                              (patch_right <= bbrights) &\
                              (bottom_on_patch)

        # Two sides are on the patch
        leftrightside_on_patch = (left_on_patch) &\
                                 (right_on_patch) &\
                                 (bbtops <= patch_top) &\
                                 (patch_bottom <= bbbottoms)
        topbottomside_on_patch = (bblefts <= patch_left) &\
                                 (patch_right <= bbrights) &\
                                 (top_on_patch) &\
                                 (bottom_on_patch)

        idx_of_bb_on_patch = leftside_on_patch | rightside_on_patch | \
            topside_on_patch | bottomside_on_patch | \
            leftrightside_on_patch | topbottomside_on_patch
        idx_of_bb_on_patch = np.where(idx_of_bb_on_patch)[0]
        idx_of_bb_on_patch = list(set(idx_of_bb_on_patch))
        return list(idx_of_bb_on_patch)

    def annotation_cover_patch(self, coords, x, y):
        """Check if the annotation is covering the whole patch.

        Args:
            coords (np.array) : Coordinations of annotations.
            px (int) : X coordinate of left top corner of the patch.
            py (int) : Y coordinate of left top corner of the patch.

        Returns:
            idx_of_bb_on_patch (list): List of np.int64s which are the indices
                of bounding boxes on the patch.
        """
        bblefts = np.min(coords, axis=1)[:, 0]
        bbtops = np.min(coords, axis=1)[:, 1]
        bbrights = np.max(coords, axis=1)[:, 0]
        bbbottoms = np.max(coords, axis=1)[:, 1]

        patch_left = x
        patch_right = x + self.p_width
        patch_top = y
        patch_bottom = y + self.p_height

        idx_of_bb_on_patch = (bblefts <= patch_left) &\
                             (patch_right <= bbrights) &\
                             (bbtops <= patch_top) &\
                             (patch_bottom <= bbbottoms)
        idx_of_bb_on_patch = np.where(idx_of_bb_on_patch)[0]
        return list(idx_of_bb_on_patch)

    def to_bb(self, coord):
        """Convert coordinates to voc coordinates.

        Args:
            coord (list): List of coordinates stored as below::

                [[xOfOneCorner, yOfOneCorner],
                 [xOfApex,      yOfApex]]

        Returns:
            outer_coord (list): List of coordinates stored as below::

                [[xmin, ymin],
                 [xmin, ymax],
                 [xmax, ymax],
                 [xmax, ymin]]

        """
        coord = np.array(coord)
        xmin, ymin = np.min(coord, axis=0)
        xmax, ymax = np.max(coord, axis=0)
        outer_coord = [[xmin, ymin],
                       [xmin, ymax],
                       [xmax, ymax],
                       [xmax, ymin]]
        return outer_coord

    def find_masks(self, x, y, cls):
        """Get the masked area corresponding to the given patch area.

        Args:
            x (int): X-axis offset of a patch.
            y (int): Y-axis offset of a patch.
            cls (str): Class of the patch or the bounding box or the segmented
                area.

        Returns:
            masks (list): List containing a dict of coords and its class. This
                coords is a path to the png image.
        """
        if not self.patch_on_annotation(cls, x, y):
            return []
        else:
            # Find mask coords
            patch_mask = self.annotation.get_patch_mask(
                cls, x, y, self.p_width, self.p_height)
            mask_path = "{}/{}/masks/{}/{:06}_{:06}.{}".format(
                self.save_to, self.filestem, cls, x, y, self.ext)
            if not self.no_patches:
                cv2.imwrite(mask_path, patch_mask, (cv2.IMWRITE_PXM_BINARY, 1))
            # contours, _ = cv2.findContours(
            #   patch_mask, cv2.RETR_EXTERNAL, cv2.CHAIN_APPROX_NONE)
            masks = []
            mask = {"coords": mask_path, "class": cls}
            masks.append(mask)
            return masks

    def remove_dup_in_results(self):
        """Remove duplicate results in self.result["results"]
        """
        results = set()
        for result in self.result["result"]:
            results.add(json.dumps(result))
        self.result["result"] = [json.loads(res) for res in list(results)]

    def save_results(self):
        """Save the extraction results.

        Saves some metadata with the patches results.

        """
        self.result["slide"] = self.slide.path
        self.result["method"] = self.method
        self.result["wsi_width"] = self.wsi_width
        self.result["wsi_height"] = self.wsi_height
        self.result["patch_width"] = self.p_width
        self.result["patch_height"] = self.p_height
        self.result["overlap_width"] = self.o_width
        self.result["overlap_hegiht"] = self.o_height
        self.result["offset_x"] = self.offset_x
        self.result["offset_y"] = self.offset_y
        self.result["ext"] = self.ext
        self.result["start_sample"] = self.start_sample
        self.result["finished_sample"] = self.finished_sample
        self.result["no_patches"] = self.no_patches
        self.result["on_foreground"] = self.on_foreground
        self.result["on_annotation"] = self.on_annotation
        self.result["dot_bbox_width"] = self.dot_bbox_width
        self.result["dot_bbox_height"] = self.dot_bbox_height
        self.result["magnification"] = self.magnification
        self.result["dryrun"] = self.dryrun
        self.result["save_to"] = str(Path(self.save_to).absolute())
        self.result["classes"] = sorted(self.classes)

        self.remove_dup_in_results()

        with open(
            "{}/{}/results.json".format(self.save_to, self.filestem),
                "w") as f:
            json.dump(self.result, f, indent=4)

        coords = pd.DataFrame(self.result["result"])
        if not self.result["result"]:
            return

        coords.sort_values(by=["x", "y"], inplace=True)
        coords.reset_index(drop=True, inplace=True)
        if self.method == "classification":
            for cls in self.classes:
                coords[cls] = (coords["class"] == cls)

            # aggregate if a patch has multiple classes
            if coords.duplicated(subset=["x", "y", "w", "h"]).sum() > 0:
                dup_to_keep = ~coords.duplicated(
                    subset=["x", "y", "w", "h"], keep="first")

                for i, row in coords[dup_to_keep].iterrows():
                    dup_x = coords.x == row.x
                    dup_y = coords.y == row.y
                    dup_w = coords.w == row.w
                    dup_h = coords.h == row.h
                    dup = coords[dup_x & dup_y & dup_w & dup_h]
                    coords.loc[i, self.classes] = dup[self.classes].any()

                coords.drop_duplicates(
                    subset=["x", "y", "w", "h"], keep="first", inplace=True)

            coords.drop(columns="class", inplace=True)

        elif self.method == "detection":
            # column name: bbs
            # not implemented
            pass

        elif self.method == "segmentation":
            # column name: masks
            for cls in self.classes:
                coords[cls] = False

            def mask_cls_to_column(x):
                for mask in x.masks:
                    coords.loc[x.name, mask["class"]] = True

            coords.apply(mask_cls_to_column, axis=1)

        coords.to_csv(
            "{}/{}/coords.csv".format(self.save_to, self.filestem),
            index=None)

    def get_patch(self, x, y, classes=False):
        """Extract a single patch.

        Args:
            x (int): X-axis offset of a patch.
            y (int): Y-axis offset of a patch.
            classes (list): For the case of method is classification, extract
                the patch for multiple times if the patch is on the border of
                two or more classes. To prevent patcher to extract a single
                patch for multiple classes, `on_annotation=1.0` should work.
        """
        if self.on_foreground:
            if not self.patch_on_foreground(x, y):
                return
        if self.on_annotation:
            on_annotation_classes = []
            for cls in classes:
                if self.patch_on_annotation(cls, x, y):
                    on_annotation_classes.append(cls)
        else:
            on_annotation_classes = ["foreground"]
        for cls in on_annotation_classes:
            if not self.no_patches:
                patch = self.crop_patch(x, y)
                self.save_patch(patch, "{}/{}/patches/{}/{:06}_{:06}.{}".format(
                    self.save_to, self.filestem, cls, x, y, self.ext))
            self.save_patch_result(x, y, cls)

    def get_patch_parallel(self, classes=False, max_workers=-1):
        """Run get_patch() in parallel.

        Args:
            classes (list): Classes to extract.
            max_workers (int): Workers to run. -1 runs with cores*5 threads.
        """
        for cls in classes:
            assert cls in self.on_annotation, f"on_annotation of {cls} not set"

        if max_workers == 0 | max_workers < -1:
            msg = "max_workers must be 1 or larger, or -1"
            msg += f", got {max_workers}"
            warnings.warn(msg)
        for cls in classes:
            if not self.no_patches:
                self.verify.make_dir(
                    "{}/{}/patches/{}".format(
                        self.save_to, self.filestem, cls))
                if self.method == "segmentation":
                    self.verify.make_dir(
                        "{}/{}/masks/{}".format(
                            self.save_to, self.filestem, cls))

        if self.start_sample:
            self.get_random_sample("start", 3)

        if max_workers == -1:
            max_workers = os.cpu_count()*5
        else:
            max_workers = max_workers

        with ThreadPoolExecutor(max_workers=max_workers) as executor:
            xs = [iter[0] for iter in self.iterator]
            ys = [iter[1] for iter in self.iterator]
            if self.verbose:
                desc = f"[{self.filepath} {self.p_width}x{self.p_height}]"
                list(tqdm(executor.map(
                    self.get_patch, xs, ys,
                    [classes for _ in self.iterator]
                ), desc=desc, total=len(self.iterator)))
            else:
                executor.map(
                    self.get_patch, xs, ys,
                    [classes for _ in self.iterator]
                )

        # save results
        self.save_results()

        if self.finished_sample:
            self.get_random_sample("finished", 3)

    def get_mini_patch_parallel(self, classes=False):
        for cls in classes:
            self.verify.make_dir(
                "{}/{}/mini_patches/{}".format(
                    self.save_to, self.filestem, cls))

        for patch in self.result["result"]:
            for bb in patch["bbs"]:
                if bb["class"] not in classes:
                    continue
                # not properly works with the magnification settings
                mini_patch = self.crop_patch(
                    patch["x"] + bb["x"],
                    patch["y"] + bb["y"],
                    bb["w"],
                    bb["h"]
                )
                self.save_patch(
                    mini_patch,
                    "{}/{}/mini_patches/{}/{:06}_{:06}.{}".format(
                        self.save_to, self.filestem,
                        bb["class"],
                        patch["x"] + bb["x"],
                        patch["y"] + bb["y"],
                        self.ext))

    def patch_on_foreground(self, x, y):
        """Check if the patch is on the foreground area.

        Args:
            x (int): X-axis offset of a patch.
            y (int): Y-axis offset of a patch.

        Returns:
            (bool): Whether the patch is on the foreground area.
        """
        patch_mask = self.annotation.get_patch_mask(
            "foreground", x, y, self.p_width, self.p_height)
        return (patch_mask.sum() / self.p_area) >= self.on_foreground

    def patch_on_annotation(self, cls, x, y):
        """Check if the patch is on the annotation area of a class.

        Args:
            cls (str): Class of the patch or the bounding box or the segmented
                area.
            x (int): X-axis offset of a patch.
            y (int): Y-axis offset of a patch.

        Returns:
            (bool): Whether the patch is on the anntation.
        """
        patch_mask = self.annotation.get_patch_mask(
            cls, x, y, self.p_width, self.p_height)
        return (patch_mask.sum() / self.p_area) >= self.on_annotation[cls]

    def get_random_sample(self, phase, sample_count=1):
        """Get random patch to check if the patcher can work properly.

        Args:
            phase (str): When to check. One of {start, finish}
            sample_count (int): Number of patches to extract.
        """
        for i in range(sample_count):
            x = random.choice(self.x_lefttop)
            y = random.choice(self.y_lefttop)
            patch = self.crop_patch(x, y)
            self.save_patch(patch, "{}/{}/{}_sample/{:06}_{:06}.{}".format(
                self.save_to, self.filestem, phase, x, y, self.ext))

    def crop_patch(self, x, y, w=False, h=False):
        if not w:
            w = self.p_width
        if not h:
            h = self.p_height
        return self.slide.crop(x, y, w, h)

    def save_patch(self, patch, save_as):
        if patch.mode == "RGBA" and self.ext == "jpg":
            warnings.warn(
                "patch has RGBA data. Discarding alpha to save as jpg")
            patch = patch.convert("RGB")

        if self.magnification:
            patch = patch.resize((
                int(self.p_width//self.p_scale),
                int(self.p_height//self.p_scale)
            ))

        patch.save(save_as)
